import numpy as np
import matplotlib.pyplot as plt
from pyinterpolate.distance.calculate_distances import calc_block_to_block_distance
from pyinterpolate.semivariance.semivariogram_estimation.calculate_semivariance import calculate_semivariance
from pyinterpolate.semivariance.semivariogram_fit.fit_semivariance import TheoreticalSemivariogram

from pyinterpolate.semivariance.areal_semivariance.within_block_semivariance.calculate_semivariance_within_blocks\
    import calculate_semivariance_within_blocks
from pyinterpolate.semivariance.areal_semivariance.within_block_semivariance.calculate_average_semivariance\
    import calculate_average_semivariance
from pyinterpolate.semivariance.areal_semivariance.block_to_block_semivariance.\
    calculate_block_to_block_semivariance import calculate_block_to_block_semivariance


class ArealSemivariance:
    """
    Class calculates semivariance of areas for Poisson Kriging (area to area and area to point).

    INITIALIZATION PARAMS:

    :param areal_data: (numpy array / list) [area_id, area_geometry, centroid x,
        centroid y, value],
    :param max_areal_range: (float) max distance to perform distance and semivariance calculations,
    :param areal_step_size: (float) step size for search radius,
    :param areal_points_data: (numpy array / list of lists)
        [area_id, [point_position_x, point_position_y, value]]
    :param weighted_semivariance: (bool) if False then each distance is treated equally when calculating
        theoretical semivariance; if True then semivariances closer to the point of origin have more weight,
    :param verbose: (bool) if True then all messages are printed, otherwise nothing.
    """

    def __init__(self, areal_data, areal_step_size, max_areal_range,
                 areal_points_data, weighted_semivariance=False, verbose=False):

        """
        INPUT:

        :param areal_data: (numpy array / list of lists)
            [area_id, area_geometry, centroid coordinate x, centroid coordinate y, value],
        :param areal_step_size: (float) step size for search radius,
        :param max_areal_range: (float) max distance to perform distance and semivariance calculations,
        :param areal_points_data: (numpy array / list of lists)
            [area_id, [point_position_x, point_position_y, value]]
        :param weighted_semivariance: (bool) if False then each distance is treated equally when calculating
            theoretical semivariance; if True then semivariances closer to the point have more weight,
        :param verbose: (bool) if True then all messages are printed, otherwise nothing.
        """

        # Initial data
        self.areal_dataset = areal_data
        self.areal_centroids = self.areal_dataset[:, 2:]
        self.areal_lags = np.arange(0, max_areal_range, areal_step_size)
        self.areal_ss = areal_step_size
        self.areal_max_range = max_areal_range
        self.within_area_points = areal_points_data
        self.weighted_semivariance = weighted_semivariance

        # Semivariogram models
        self.experimental_semivariogram = None
        self.theoretical_semivariance_model = None
        self.inblock_semivariance = None
        self.within_block_semivariogram = None
        self.between_blocks_semivariogram = None
        self.regularized_semivariogram = None

        # Model parameters
        self.distances_between_blocks = None
        self.verbose = verbose

    def _calculate_empirical_semivariance(self):
        """Method calculates theoretical semivariance between areal centroids
        :return gamma: (float) semivariance between areal centroids."""
        gamma = calculate_semivariance(self.areal_centroids,  self.areal_ss, self.areal_max_range)
        return gamma

    def _calculate_theoretical_semivariance(self):
        """Method calculates theoretical semivariogram of a dataset
        :return ts: (TheoreticalSemivariogram) Fitted semivariogram model."""
        ts = TheoreticalSemivariogram(self.areal_centroids, self.experimental_semivariogram)
        ts.find_optimal_model(weighted=self.weighted_semivariance)
        return ts

    # ---------------------------------- AREAL SEMIVARIANCE CALCULATIONS ----------------------------------

    def regularize_semivariogram(self, within_block_semivariogram=None, between_blocks_semivariogram=None,
                                 empirical_semivariance=None, theoretical_semivariance_model=None):
        """
        Function calculates regularized point support semivariogram in the form given in:

        Goovaerts P., Kriging and Semivariogram Deconvolution in the Presence of Irregular Geographical Units,
            Mathematical Geology 40(1), 101-128, 2008

        Function has the form: gamma_v(h) = gamma(v, v_h) - gamma_h(v, v) where:

        - gamma_v(h) - regularized semivariogram,
        - gamma(v, v_h) - semivariogram value between any two blocks separated by the distance h,
        - gamma_h(v, v) - arithmetical average of within-block semivariogram

        INPUT:

        :param within_block_semivariogram: (numpy array) mean semivariance between the blocks:
            yh(v, v) = 1 / (2*N(h)) SUM(from a=1 to N(h)) [y(va, va) + y(va+h, va+h)], where:
            y(va, va) and y(va+h, va+h) are the inblock semivariances of block a and block a+h separated
                by the distance h weighted by the inblock population.
        :param between_blocks_semivariogram: (numpy array) semivariance between all blocks calculated from the
            theoretical model,
        :param empirical_semivariance: (numpy array) empirical semivariance between area centroids, default=None, if
            None is provided then empirical semivariance is computed by the _calculate_empirical_semivariance
            method from area centroids,
        :param theoretical_semivariance_model: (TheoreticalSemivariogram) theoretical semivariance model from
            TheoreticalSemivariance class, default is None, if None is provided then theoretical model is derived
            from area centroids and empirical semivariance.

        OUTPUT:

        :return: semivariance: numpy array of pair of lag and semivariance values where:

            - semivariance[0] = array of lags
            - semivariance[1] = array of lag's values
            - semivariance[2] = array of number of points in each lag
        """

        # SET SEMIVARIANCES
        if empirical_semivariance is None:
            self.experimental_semivariogram = self._calculate_empirical_semivariance()
        else:
            self.experimental_semivariogram = empirical_semivariance

        if theoretical_semivariance_model is None:
            self.theoretical_semivariance_model = self._calculate_theoretical_semivariance()
        else:
            self.theoretical_semivariance_model = theoretical_semivariance_model

        # GET WITHIN-AREA SEMIVARIANCE gamma_h(v, v)

        if not within_block_semivariogram:
            self.within_block_semivariogram = self.calculate_mean_semivariance_between_blocks()
        else:
            self.within_block_semivariogram = within_block_semivariogram

        # GET SEMIVARIANCE BETWEEN BLOCKS gamma(v, v_h)

        if not between_blocks_semivariogram:
            self.between_blocks_semivariogram = self.calculate_semivariance_between_areas()
        else:
            self.between_blocks_semivariogram = between_blocks_semivariogram

        regularized_semivariogram = self.between_blocks_semivariogram.copy()
        regularized_semivariogram[:, 1] = (
            self.between_blocks_semivariogram[:, 1] - self.within_block_semivariogram[:, 1]
        )

        # CHECK VALUES BELOW 0

        for idx, row in enumerate(regularized_semivariogram):
            if row[1] < 0:
                regularized_semivariogram[idx, 1] = 0

        self.regularized_semivariogram = regularized_semivariogram.copy()

        return self.regularized_semivariogram

    # ---------------------------------- WITHIN-BLOCK SEMIVARIANCE CALCULATIONS ----------------------------------

    def calculate_mean_semivariance_between_blocks(self, distances=None):
        """
        Function calculates average semivariance between blocks separated by a vector h according to the equation:

        yh(v, v) = 1 / (2*N(h)) SUM(from a=1 to N(h)) [y(va, va) + y(va+h, va+h)], where:

        - y(va, va) and y(va+h, va+h) are estimated according to the function calculate_semivariance_within_blocks,
        - h are estimated according to the block_to_block_distances function.

        INPUT:

        :param distances: if given then this step of calculation is skipped

        OUTPUT:

        :return: [s, d]

                - s - semivariances in the form: list of [[lag, semivariance], [lag_x, semivariance_x], [..., ...]]
                - if distances: d - distances between blocks (dict) in the form: {area_id: {other_area_id: distance,
                    other_area_id: distance,}}
                - else: d = 0.
        """

        # Calculate inblock semivariance
        if self.verbose:
            print('Start of the inblock semivariance calculation')
        # Pass numpy array with [area id, [points within area and their values]] and semivariogram model
        self.inblock_semivariance = np.array(calculate_semivariance_within_blocks(self.within_area_points,
                                                                                  self.theoretical_semivariance_model))
        if self.verbose:
            print('Inblock semivariance calculated successfully')

        # Calculate distance between blocks
        if distances is None:
            if self.verbose:
                print('Distances between blocks: calculation starts')
            self.distances_between_blocks = calc_block_to_block_distance(self.within_area_points)
            if self.verbose:
                print('Distances between blocks have been calculated')
        else:
            if self.verbose:
                print('Distances between blocks are provided, distance skipped, model parameters updated')
            self.distances_between_blocks = distances

        # Calc average semivariance
        avg_semivariance = calculate_average_semivariance(self.distances_between_blocks, self.inblock_semivariance,
                                                          self.areal_lags, self.areal_ss)
        return avg_semivariance

    # ---------------------------------- BLOCK TO BLOCK SEMIVARIANCE CALCULATIONS ----------------------------------

    def calculate_semivariance_between_areas(self):
        """
        Function calculates semivariance between areas based on their division into smaller blocks. It is
            gamma(v, v_h) - semivariogram value between any two blocks separated by the distance h.

        OUTPUT:

        :return semivariogram: (array) Semivariances between blocks for a given lags based on the inblock division into
            smaller blocks. It is numpy array of the form [[lag, semivariance], [next lag, other semivariance], [...]].
        """
        # Check if distances are calculated
        if self.distances_between_blocks is None:
            if self.verbose:
                print('Distances between blocks: calculation starts')
            self.distances_between_blocks = calc_block_to_block_distance(self.within_area_points)
            if self.verbose:
                print('Distances between blocks have been calculated')

        if self.verbose:
            print('Calculation of semivariances between areas separated by chosen lags')
        blocks = calculate_block_to_block_semivariance(self.within_area_points,
                                                       self.distances_between_blocks,
                                                       self.theoretical_semivariance_model)
        if self.verbose:
            print('Semivariance between blocks for a given lags calculated')
            print('Calculation of the mean semivariance for a given lag')
        semivariogram = []

        for areal_lag in self.areal_lags:
            dists_and_semivars = blocks[0].copy().flatten()
            smvs = []
            for dists_sem_group in dists_and_semivars:
                dists_sem_group = np.array(dists_sem_group)

                # Get dists greater than areal lag minus step size
                lower_limit = areal_lag - self.areal_ss
                upper_limit = areal_lag + self.areal_ss
                chosen_dists_upper_limit = dists_sem_group[dists_sem_group[:, 0] > lower_limit]

                # Get dists lower than areal lag plus step size
                chosen_dists = chosen_dists_upper_limit[chosen_dists_upper_limit[:, 0] < upper_limit]

                if len(chosen_dists[:, 1]) != 0:
                    semivars_inside = np.mean(chosen_dists[:, 1])
                else:
                    semivars_inside = 0
                smvs.append(semivars_inside)

            if np.sum(smvs) == 0:
                semivariogram.append([areal_lag, 0])
            else:
                smvs = np.mean(smvs)
                semivariogram.append([areal_lag, smvs])

        if self.verbose:
            print('End of block to block semivariogram calculation')
        return np.array(semivariogram)

    def show_semivariograms(self):
        """
        Function shows semivariograms calculated by the class: Empirical semivariogram, Theoretical model,
        Inblock Semivariance, Within-block semivariogram, Between blocks semivariogram, Regularized output.
        """
        plt.figure(figsize=(12, 12))
        plt.plot(self.experimental_semivariogram[:, 0], self.experimental_semivariogram[:, 1],
                 color='blue', marker='o')
        plt.plot(self.experimental_semivariogram[:, 0],
                 self.theoretical_semivariance_model.predict(self.experimental_semivariogram[:, 0]),
                 color='black', linestyle='dotted')
        mean_inblock_semivar = np.mean(self.inblock_semivariance[:, 1])
        inblock_sem = np.ones((1, len(self.experimental_semivariogram[:, 0]))) * mean_inblock_semivar
        plt.plot(self.experimental_semivariogram[:, 0], inblock_sem[0],
                 color='yellow', linestyle='-')
        plt.plot(self.within_block_semivariogram[:, 0], self.within_block_semivariogram[:, 1], color='green')
        plt.plot(self.between_blocks_semivariogram[:, 0], self.between_blocks_semivariogram[:, 1], color='pink')
        plt.plot(self.regularized_semivariogram[:, 0], self.regularized_semivariogram[:, 1], color='red')

        plt.legend(['Empirical semivariogram', 'Theoretical model', 'Inblock Semivariance',
                    'Within-block semivariogram', 'Between blocks semivariogram', 'Regularized output'])
        plt.title('Areal semivariograms')
        plt.xlabel('Distance')
        plt.ylabel('Semivariance')
        plt.show()



import numpy as np

from pyinterpolate.transform.prepare_kriging_data import prepare_kriging_data
from pyinterpolate.distance.calculate_distances import calc_point_to_point_distance


class Krige:
    """
    Class for kriging interpolation of the unknown values at a given location (position). Class takes two arguments
    during the initialization:
    semivariogram_model - semivariogram model,
    known_points - array of known values [x, y, val]

    Available methods:

    - ordinary_kriging - ordinary kriging of unknown point value,
    - simple_kriging - simple kriging of unknown point value.

    Class methods may raise ValueError if estimated value is below 0. You may use try: ... except: ... statement to overwrite those values
        with some constant or NaN or you could use different semivariogram model. Sometimes this problem is related to
        the input data, especially to the clustered groups of points. In this case aggregate those clusters and then estimate
        semivariogram and perform kriging.

    INITLIALIZATION PARAMS:

    :param semivariogram_model: semivariogram model returned by TheoreticalSemivariogram class
    :param known_points: dataset with known values and locations. Each column should represent different dimension and
        the last column represents values example: [[dim_x1, dim_y1, val_1], [dim_x2, dim_y2, val_2]].
    """

    def __init__(self, semivariogram_model, known_points):
        """
        INPUT:

        :param semivariogram_model: (TheoreticalSemivariogram) Theoretical Semivariogram used for data interpolation,
        :param known_points: (numpy array) dataset with known values and locations. Each column should represent different dimension and the last column represents values
        example: [[dim_x1, dim_y1, val_1], [dim_x2, dim_y2, val_2]]
        """

        self.dataset = known_points
        self.model = semivariogram_model
        self.distances = None

    def ordinary_kriging(self, unknown_location, number_of_neighbours, test_anomalies=True):
        """
        Function predicts value at unknown location with Ordinary Kriging technique.

        INPUT:

        :param unknown_location: (tuple) position of unknown location,
        :param number_of_neighbours: (int) number of the closest locations to the unknown position which should be
            included in the modeling,
        :param test_anomalies: (bool) check if weights are negative.

        OUTPUT:

        :return: predicted, error, estimated mean, weights
            [value_in_unknown_location, error, estimated_mean, weights]
        """

        prepared_data = prepare_kriging_data(unknown_position=unknown_location,
                                             data_array=self.dataset,
                                             number_of_neighbours=number_of_neighbours)
        n = number_of_neighbours
        unknown_distances = prepared_data[:, -1]
        k = self.model.predict(unknown_distances)
        k = k.T
        k_ones = np.ones(1)[0]
        k = np.r_[k, k_ones]

        dists = calc_point_to_point_distance(prepared_data[:, :-2])

        predicted_weights = self.model.predict(dists.ravel())
        predicted = np.array(predicted_weights.reshape(n, n))
        p_ones = np.ones((predicted.shape[0], 1))
        predicted_with_ones_col = np.c_[predicted, p_ones]
        p_ones_row = np.ones((1, predicted_with_ones_col.shape[1]))
        p_ones_row[0][-1] = 0.
        weights = np.r_[predicted_with_ones_col, p_ones_row]

        w = np.linalg.solve(weights, k)
        zhat = prepared_data[:, -2].dot(w[:-1])

        # Test for anomalies
        if test_anomalies:
            if zhat < 0:
                user_input_message = 'Estimated value is below zero and it is: {}. \n'.format(zhat)
                text_error = user_input_message + 'Program is terminated. Try different semivariogram model. ' \
                                                  '(Did you use gaussian model? \
                            If so then try to use other models like linear or exponential) and/or analyze your data \
                            for any clusters which may affect the final estimation'

                raise ValueError(text_error)

        sigmasq = (w.T * k)[0]
        if sigmasq < 0:
            sigma = 0
        else:
            sigma = np.sqrt(sigmasq)
        return zhat, sigma, w[-1], w

    def simple_kriging(self, unknown_location, number_of_neighbours, mu=None, test_anomalies=True):
        """
        Function predicts value at unknown location with Simple Kriging technique.

        INPUT:

        :param unknown_location: (tuple) position of unknown location,
        :param number_of_neighbours: (int) number of the closest locations to the unknown position which should be
            included in the modeling,
        :param mu: (float) global mean which should be known before processing. If not given then it is calculated
            from the sample but then it may cause a relative large errors (this mean is expectation of the random field,
            so without knowledge of the ongoing processes it is unknown),
        :param test_anomalies: (bool) check if weights are negative.

        OUTPUT:

        :return: predicted, error, mean, weights:
            [value_in_unknown_location, error, mean, weights]
        """

        prepared_data = prepare_kriging_data(unknown_position=unknown_location,
                                             data_array=self.dataset,
                                             number_of_neighbours=number_of_neighbours)
        n = number_of_neighbours

        if mu is None:
            vals = self.dataset[:, -1]
            mu = np.sum(vals)
            mu = mu / len(vals)

        unknown_distances = prepared_data[:, -1]
        k = self.model.predict(unknown_distances)
        k = k.T

        dists = calc_point_to_point_distance(prepared_data[:, :-2])
        predicted_weights = self.model.predict(dists.ravel())
        predicted = np.array(predicted_weights.reshape(n, n))

        w = np.linalg.solve(predicted, k)
        r = prepared_data[:, -2] - mu
        zhat = r.dot(w)
        zhat = zhat + mu

        # Test for anomalies
        if test_anomalies:
            if zhat < 0:
                user_input_message = 'Estimated value is below zero and it is: {}. \n'.format(zhat)
                text_error = user_input_message + 'Program is terminated. Try different semivariogram model. ' \
                                                  '(Did you use gaussian model? \
                            If so then try to use other models like linear or exponential) and/or analyze your data \
                            for any clusters which may affect the final estimation'

                raise ValueError(text_error)

        sigmasq = (w.T * k)[0]
        if sigmasq < 0:
            sigma = 0
        else:
            sigma = np.sqrt(sigmasq)
        return zhat, sigma, mu, w



import unittest
import numpy as np
from pyinterpolate.distance.calculate_distances import calc_block_to_block_distance


class TestCalcBlock2BlockDistance(unittest.TestCase):

    def test_calc_block_to_block_distance(self):
        # Test block_to_block_distances

        areal_test_data = [
            ['id0', [[0, 1, 1],
                     [0, 2, 1],
                     [0, 3, 1]]
             ],
            ['id1', [[1, 0, 2],
                     [1, 1, 2]]],

            ['id2', [[2, 1, 3],
                     [2, 2, 4],
                     [2, 3, 4]]]
        ]

        x = calc_block_to_block_distance(areal_test_data)
        self.assertEqual(int(np.sum(x[0])), 12, "Block to block distance is not calculated correctly, should be 12")


if __name__ == '__main__':
    unittest.main()



import numpy as np
from pyinterpolate.distance.calculate_distances import calc_point_to_point_distance
from tqdm import tqdm


def block_pair_semivariance(block_a, block_b, semivariogram_model):
    """
    Function calculates average semivariance between two blocks based on the counts inside the block.
    :param block_a: block A points in the form of array [[point x1A, point y1A, value v1A],
                                                         [point x2A, point y2A, value v2A],
                                                         [...]
                                                         [point xnA, point ynA, value vnA]]
        All coordinates from the array must be placed inside the block!
    :param block_b: block B points in the form of array [[point x1B, point y1B, value v1B],
                                                         [point x2B, point y2B, value v2B],
                                                         [...]
                                                         [point xnB, point ynB, value vnB]]
        All coordinates from the array must be placed inside the block!
    :param semivariogram_model: (TheoreticalSemivariogram) theoretical semivariance model from TheoreticalSemivariance
        class. Model must be fitted and calculated.
    :return semivariance_mean: (float) Average semivariance between blocks divided by points.
    """
    distances_between_points = calc_point_to_point_distance(block_a, block_b).flatten()

    semivariances = []
    for dist in distances_between_points:
        semivariances.append(
            semivariogram_model.predict(dist)
        )

    semivariance_mean = np.sum(semivariances) / len(semivariances)

    return semivariance_mean


def calculate_block_to_block_semivariance(points_within_area, distances_between_blocks, semivariogram_model):
    """
    Function calculates semivariances between all blocks passed into it based on the points and their
    values inside the blocks.

    :param points_within_area: (numpy array) with area id and points and respective values inside area:
        [area id, [
                    [point x, point y, value] ...
                  ]
        ]
    :param distances_between_blocks: distances_arrays: (arrays)
        array[0] - list of distances between blocks,
        array[1] - list of blocks ids.

        array[0]: [
                    distances[id A to id A, id A to id B, id A to id N],
                    distances[id B to id A, id B to id B, id B to id N],
                    distances[id N to id A, id N to id B, id N to id N]
                ],
        array[1]: [id A, id B, id N]
    :param semivariogram_model: (TheoreticalSemivariogram) Theoretical Semivariogram object,
    :return output_array: (numpy array) semivariances and distances array:
        output_array[0] - list of distances and semivariances between blocks,
        output_array[1] - list of blocks ids.

        output_array[0]: [[[distance between A and A, semivariance between A and A],
                          [distanace between A and B, semivariance between A and B]],

                          [[distance between B and A, semivariance between B and A],
                          [distanace between B and B, semivariance between B and B]],

                          ...]
        output_array[1]: [id A, id B, id ...]
    """
    blocks_ids = distances_between_blocks[1]

    if isinstance(points_within_area, list):
        points_within_area = np.array(points_within_area)

    block_pairs_smvs = dict()

    for first_idx, first_block_id in enumerate(tqdm(blocks_ids)):
        first_block_points = points_within_area[points_within_area[:, 0] == first_block_id]

        for second_idx, second_block_id in enumerate(blocks_ids):

            pair = (first_idx, second_idx)

            if first_block_id == second_block_id:
                block_pairs_smvs[pair] = [0, 0]
            else:

                # CHECK IF SEMIVAR IS NOT ACTUALLY CALCULATED

                rev_pair = (second_idx, first_idx)

                if (rev_pair in block_pairs_smvs.keys()) and (pair not in block_pairs_smvs.keys()):
                    block_pairs_smvs[pair] = block_pairs_smvs[rev_pair]
                else:
                    # Select distance from the first selected block to the second selected block
                    distance = distances_between_blocks[0][first_idx, second_idx]

                    # Select coordinates of the block centroids
                    second_block_points = points_within_area[points_within_area[:, 0] == second_block_id]

                    # Calculate semivariance between blocks
                    semivariance = block_pair_semivariance(first_block_points[0][1], second_block_points[0][1],
                                                           semivariogram_model)
                    block_pairs_smvs[pair] = [distance, semivariance]

    # Generate output array
    block_to_block_semivariance = []
    for fidx, block_id in enumerate(blocks_ids):
        semivars = []
        for sidx, block_id_2 in enumerate(blocks_ids):
            semivars.append(block_pairs_smvs[(fidx, sidx)])
        block_to_block_semivariance.append(semivars)

    output_array = np.array([block_to_block_semivariance, blocks_ids])
    return output_array



import unittest
import os
import numpy as np
from pyinterpolate.io_ops import prepare_areal_shapefile


class TestPrepareArealShapefile(unittest.TestCase):

    def test_prepare_areal_shapefile(self):

        my_dir = os.path.dirname(__file__)
        path_to_areal_file = os.path.join(my_dir, '../sample_data/test_areas_pyinterpolate.shp')

        # Read without id column and without value column
        try:
            minimal_dataset = prepare_areal_shapefile(path_to_areal_file)
        except TypeError:
            assert True
        else:
            assert False

        minimal_dataset = prepare_areal_shapefile(path_to_areal_file, value_column_name='value')

        # Tests:
        # Last column values are not nan
        isnan_test = ~np.any(np.isnan(minimal_dataset[:, -1].astype(np.float)))
        self.assertTrue(isnan_test, "NaN values in last column")

        # Read with id column
        dataset_with_id = prepare_areal_shapefile(path_to_areal_file, id_column_name='id', dropnans=False)

        # Tests:
        # Must have 5 columns
        test_cols_dataset = len(dataset_with_id[0]) == 5
        self.assertTrue(test_cols_dataset, "Dataset should have 5 columns")


if __name__ == '__main__':
    unittest.main()



import unittest
import os
import numpy as np
import geopandas as gpd
from pyinterpolate.io_ops.get_points_within_area import get_points_within_area
from pyinterpolate.io_ops.prepare_areal_shapefile import prepare_areal_shapefile
from pyinterpolate.semivariance.areal_semivariance.areal_semivariance import ArealSemivariance

from pyinterpolate.semivariance.areal_semivariance.block_to_block_semivariance.calculate_block_to_block_semivariance\
    import calculate_block_to_block_semivariance


class TestCalculateBlock2BlockSemivariance(unittest.TestCase):

    def test_calculate_block_to_block_semivariance(self):
        # Data prepration
        my_dir = os.path.dirname(__file__)

        areal_dataset = os.path.join(my_dir, '../sample_data/test_areas_pyinterpolate.shp')
        subset = os.path.join(my_dir, '../sample_data/test_points_pyinterpolate.shp')

        a_id = 'id'
        areal_val = 'value'
        points_val = 'value'

        # Get maximum range and set step size

        gdf = gpd.read_file(areal_dataset)

        total_bounds = gdf.geometry.total_bounds
        total_bounds_x = np.abs(total_bounds[2] - total_bounds[0])
        total_bounds_y = np.abs(total_bounds[3] - total_bounds[1])

        max_range = min(total_bounds_x, total_bounds_y)
        step_size = max_range / 4

        areal_data_prepared = prepare_areal_shapefile(areal_dataset, a_id, areal_val)
        points_in_area = get_points_within_area(areal_dataset, subset, areal_id_col_name=a_id,
                                                points_val_col_name=points_val)

        # Set areal semivariance class
        areal_semivariance = ArealSemivariance(areal_data_prepared, step_size, max_range, points_in_area)
        areal_semivariance.regularize_semivariogram()
        blocks = calculate_block_to_block_semivariance(areal_semivariance.within_area_points,
                                                       areal_semivariance.distances_between_blocks,
                                                       areal_semivariance.theoretical_semivariance_model)
        test_block = np.array(blocks[0][0])
        mean_semivariance = int(np.mean(test_block[:, 1]))
        self.assertEqual(mean_semivariance, 212, "Average semivariance should be 212 (decimal part)")


if __name__ == '__main__':
    unittest.main()



import unittest
import numpy as np
from pyinterpolate.distance.calculate_distances import calc_point_to_point_distance


class TestCalcPoint2PointDistance(unittest.TestCase):

  def test_calc_point_to_point_distance(self):
      coords = [(35.0456, -85.2672),
                (35.1174, -89.9711),
                (35.9728, -83.9422),
                (36.1667, -86.7833)]

      d = calc_point_to_point_distance(coords)

      test_arr = np.array([[0., 4.70444794, 1.6171966, 1.88558331],
                           [4.70444794, 0., 6.0892811, 3.35605413],
                           [1.6171966, 6.0892811, 0., 2.84770898],
                           [1.88558331, 3.35605413, 2.84770898, 0.]])

      sum_d = np.sum(d.astype(int))
      sum_test_arr = np.sum(test_arr.astype(int))
      self.assertEqual(sum_d, sum_test_arr, "Distances between points are not calculated correctly")


if __name__ == '__main__':
    unittest.main()



import unittest
import os
import numpy as np
from pyinterpolate.io_ops import read_point_data


class TestReadData(unittest.TestCase):

	def test_read_data(self):
		my_dir = os.path.dirname(__file__)
		path_to_the_data = os.path.join(my_dir, '../sample_data/poland_dem_gorzow_wielkopolski')
		data = read_point_data(path_to_the_data, 'txt')

		# Check if data type is ndarray
		check_ndarray = isinstance(data, np.ndarray)
		self.assertTrue(check_ndarray, "Instance of a data type should be numpy array")

		# Check dimensions
		self.assertEqual(data.shape[1], 3, "Shape of data should be 3 - x, y, value")


if __name__ == '__main__':
	unittest.main()



import numpy as np

from pyinterpolate.distance.calculate_distances import calc_point_to_point_distance
from pyinterpolate.transform.prepare_kriging_data import prepare_poisson_kriging_data


class CentroidPoissonKriging:

    def __init__(self, semivariogram_model, known_areas, known_areas_points):
        """
        Class performs centroid-based Poisson Kriging of areas.

        :param semivariogram_model: (TheoreticalSemivariance object) Semivariogram model fitted to the
            TheoreticalSemivariance class,
        :param known_areas: (numpy array) areas in the form:
            [area_id, polygon, centroid x, centroid y, value]
        :param known_areas_points: (numpy array) array of points within areas in the form:
            [area_id, [point_position_x, point_position_y, value]],
        """

        self.model = semivariogram_model
        self.known_areas = known_areas
        self.known_areas = self.known_areas[self.known_areas[:, 0].argsort()]  # Sort by id
        self.known_areas_points = known_areas_points
        self.known_areas_points = self.known_areas_points[self.known_areas_points[:, 0].argsort()]  # Sort by id

        self.prepared_data = None

    def predict(self, unknown_area, unknown_area_points,
                number_of_neighbours, max_search_radius, weighted,
                test_anomalies=True):
        """
        Function predicts areal value in a unknown location based on the centroid-based Poisson Kriging.

        INPUT:

        :param unknown_area: (numpy array) unknown area data in the form:
            [area_id, polygon, centroid x, centroid y]
        :param unknown_area_points: (numpy array) points within an unknown area in the form:
            [area_id, [point_position_x, point_position_y, value]]
        :param number_of_neighbours: (int) minimum number of neighbours to include in the algorithm,
        :param max_search_radius: (float) maximum search radius (if number of neighbours within this search radius is
            smaller than number_of_neighbours parameter then additional neighbours are included up to
            the number_of_neighbours).
        :param weighted: (bool) distances weighted by population (True) or not (False),
        :param test_anomalies: (bool) check if weights are negative.

        OUTPUT:

        :return: prediction, error, estimated mean, weights:
            [value_in_unknown_area, error, estimated_mean, weights]
        """

        self.prepared_data = prepare_poisson_kriging_data(
            unknown_area=unknown_area, points_within_unknown_area=unknown_area_points,
            known_areas=self.known_areas, points_within_known_areas=self.known_areas_points,
            number_of_neighbours=number_of_neighbours, max_search_radius=max_search_radius,
            weighted=weighted
        )  # [id (known), coo_x, coo_y, val, dist_to_unknown, total population]

        n = self.prepared_data.shape[0]
        unknown_distances = self.prepared_data[:, -2]
        k = self.model.predict(unknown_distances)  # predicted values from distances to unknown
        k = k.T
        k_ones = np.ones(1)[0]
        k = np.r_[k, k_ones]

        data_for_distance = self.prepared_data[:, 1:3]
        dists = calc_point_to_point_distance(data_for_distance)

        predicted_weights = self.model.predict(dists.ravel())
        predicted = np.array(predicted_weights.reshape(n, n))

        # Add weights to predicted values (diagonal)

        weights_mtx = self.calculate_weight_arr()
        predicted = predicted + weights_mtx

        # Prepare weights matrix

        p_ones = np.ones((predicted.shape[0], 1))
        predicted_with_ones_col = np.c_[predicted, p_ones]
        p_ones_row = np.ones((1, predicted_with_ones_col.shape[1]))
        p_ones_row[0][-1] = 0.
        weights = np.r_[predicted_with_ones_col, p_ones_row]

        # Solve Kriging system
        try:
            w = np.linalg.solve(weights, k)
        except TypeError:
            weights = weights.astype(np.float)
            k = k.astype(np.float)
            w = np.linalg.solve(weights, k)

        zhat = self.prepared_data[:, 3].dot(w[:-1])

        # Test for anomalies
        if test_anomalies:
            if zhat < 0:
                user_input_message = 'Estimated value is below zero and it is: {}. \n'.format(zhat)
                text_error = user_input_message + 'Program is terminated. Try different semivariogram model. ' \
                                                  '(Did you use gaussian model? \
                            If so then try to use other models like linear or exponential) and/or analyze your data \
                            for any clusters which may affect the final estimation'

                raise ValueError(text_error)

        sigmasq = (w.T * k)[0]
        if sigmasq < 0:
            sigma = 0
        else:
            sigma = np.sqrt(sigmasq)
        return zhat, sigma, w[-1], w

    def calculate_weight_arr(self):
        """
        Function calculates additional weights for the diagonal matrix of predicted values.
        delta * (m' / n(u_i))
        where:
        delta is a param which indicates if this is observed area,
        m'/n(u_i) - population weighted mean of N rates
        """

        vals_of_neigh_areas = self.prepared_data[:, 3]
        pop_of_neigh_areas = self.prepared_data[:, -1]

        weighted = np.sum(vals_of_neigh_areas * pop_of_neigh_areas)
        weights_arr = weighted / np.sum(pop_of_neigh_areas)
        w = np.ones(shape=vals_of_neigh_areas.shape)
        w = (weights_arr * w)
        return np.diag(w)



from .get_areal_centroids import get_centroids
from .prepare_kriging_data import prepare_kriging_data, prepare_poisson_kriging_data, prepare_ata_data, prepare_atp_data
from .select_values_in_range import select_values_in_range
from .set_areal_weights import set_areal_weights


import numpy as np
from tqdm import tqdm
from pyinterpolate.distance.calculate_distances import calc_point_to_point_distance


def calculate_semivariance_within_blocks(points_within_area, semivariance_model):
    """
    Function calculates semivariances of points inside all given areal blocks.

    gamma(v, v) = 1/(P * P) * SUM(s->P) SUM(s'->P) gamma(u_s, u_s')
    where:
    gamma(v, v) - average within block semivariance,
    P - number of points used to discretize block v,
    u_s - point u within block v,
    gamma(u_s, u_s') - semivariance between point u_s and u_s' inside block v.

    :param points_within_area: (numpy array / list of lists) [area_id, array of points within area and their values],
    :param semivariance_model: (TheoreticalSemivariogram) Theoretical Semivariogram object,
    :return within_areas_semivariance: (numpy array) [area_id, semivariance]
    """

    within_areas_semivariance = []

    for block in tqdm(points_within_area):

        # Get area id
        area_id = block[0]

        # Calculate inblock semivariance for a given id
        number_of_points_within_block = len(block[1])  # P
        p = number_of_points_within_block * number_of_points_within_block

        distances_between_points = calc_point_to_point_distance(block[1][:, :-1])

        semivariances = semivariance_model.predict(distances_between_points)

        avg_semivariance = np.sum(semivariances) / p
        within_areas_semivariance.append([area_id, avg_semivariance])

    return within_areas_semivariance



import numpy as np
import geopandas as gpd
from shapely.geometry import Point
from pyinterpolate.kriging.areal_poisson_kriging.area_to_area.ata_poisson_kriging import AtAPoissonKriging
from pyinterpolate.kriging.areal_poisson_kriging.area_to_point.atp_poisson_kriging import AtPPoissonKriging


class ArealKriging:
    """
    Class performs kriging of areas with point support data within those polygons.

    INITIALIZATION PARAMS:

    :param semivariogram_model: (TheoreticalSemivariogram) Theoretical Semivariogram used for data interpolation,
    :param known_areas: (numpy array) areas in the form:
        [area_id, polygon, centroid x, centroid y, value]
    :param known_areas_points: (numpy array) points within areas in the form:
        [area_id, [point_position_x, point_position_y, value]],
    :param kriging_type: (str) default 'ata'; 'ata' - area to area poisson kriging,
        'atp' - area to point poisson kriging.
    """

    def __init__(self, semivariogram_model, known_areas, known_areas_points, kriging_type='ata'):
        """
        :param semivariogram_model: (Theoretical Semivariogram) Theoretical Semivariogram used for data interpolation,
        :param known_areas: (numpy array) areas in the form:
            [area_id, polygon, centroid x, centroid y, value]
        :param known_areas_points: (numpy array) array of points within areas in the form:
            [area_id, [point_position_x, point_position_y, value]],
        :param kriging_type: (string) default 'ata'.
            'ata' - area to area poisson kriging,
            'atp' - area to point poisson kriging.
        """

        kriging_types = {
            'ata': AtAPoissonKriging,
            'atp': AtPPoissonKriging
        }

        self.semivar_model = semivariogram_model
        self.areal_data_known = known_areas
        self.point_counts_within_area = known_areas_points
        self.ktype = kriging_type

        # Check kriging type

        if kriging_type not in kriging_types.keys():
            l1 = 'Provided argument is not correct. You must choose kriging type.\n'
            l2 = "'ata' - area to area poisson kriging,\n"
            l3 = "'atp' - area to point poisson kriging."
            message = l1 + l2 + l3
            raise TypeError(message)

        self.k_func = kriging_types[self.ktype](self.semivar_model,
                                                self.areal_data_known,
                                                self.point_counts_within_area)
        self.reg_model = kriging_types[self.ktype]

        self.regularized = None  # Regularized

    def predict(self, unknown_area_points, number_of_neighbours, max_search_radius):
        """
        Function predicts area value in an unknown location based on the area-to-area or area-to-point Poisson Kriging.

        INPUT:

        :param unknown_area_points: (numpy array) points within an unknown area in the form:
            [area_id, [point_position_x, point_position_y, value]],
        :param number_of_neighbours: (int) minimum number of neighbours to include in the algorithm,
        :param max_search_radius: (float) maximum search radius (if number of neighbours within this search radius is
            smaller than number_of_neighbours parameter then additional neighbours are included up to
            the number_of_neighbors).

        OUTPUT:

        :return: prediction, error, estimated mean, weights:
            [value_in_unknown_location, error, estimated_mean, weights].
        """

        prediction = self.k_func.predict(unknown_area_points,
                                         number_of_neighbours,
                                         max_search_radius)

        return prediction

    @staticmethod
    def _rmse_areal(prediction_row, u_area_val):

        est_val = prediction_row[0]
        rmse = np.sqrt((u_area_val - est_val)**2)

        prediction_vals = [
            prediction_row[0],
            prediction_row[1],
            rmse
        ]  # Estimated value, Estimated error, RMSE

        return np.array(prediction_vals)

    @staticmethod
    def _rmse_area_to_point(prediction_row, u_areal_val):
        estimated_areal_sum = np.sum(prediction_row[:, 0])
        rmse = np.sqrt((u_areal_val - estimated_areal_sum)**2)
        rmses = rmse * np.ones(prediction_row.shape[0])
        prediction_vals = np.c_[
            prediction_row[:, -1],
            prediction_row[:, :2],
            rmses
        ]  # points xy, Estimated value, Estimated error, RMSEs

        return prediction_vals

    def _get_prediction_row(self, regularized_area_id, n_neighbours, radius):

        # Divide dataset for prediction
        u_area = self.areal_data_known[self.areal_data_known[:, 0] == regularized_area_id][0]
        u_points = self.point_counts_within_area[self.point_counts_within_area[:, 0] == regularized_area_id][0]

        k_areas = self.areal_data_known[self.areal_data_known[:, 0] != regularized_area_id]
        k_points = self.point_counts_within_area[self.point_counts_within_area[:, 0] != regularized_area_id]

        # Prepare model

        temp_model = self.reg_model(
            self.semivar_model,
            k_areas,
            k_points
        )

        # Predict

        prediction = temp_model.predict(u_points,
                                        n_neighbours,
                                        radius)

        # Add RMSE of prediction

        if self.ktype == 'ata':
            output_rows = self._rmse_areal(prediction, u_area[-1])
        else:
            output_rows = self._rmse_area_to_point(prediction, u_area[-1])

        return output_rows

    def regularize_data(self, number_of_neighbours, max_search_radius, data_crs="EPSG:4326"):
        """
        Function regularizes whole dataset and creates new values and error maps based on the kriging type.
            If chosen type is area to area then function returns Geopandas GeoDataFrame with area id, areal geometry,
            estimated value, estimated prediction error, RMSE of prediction.
            If chosen type is area to point then function returns Geopandas GeoDataFrame with area id, point
                coordinates, estimated value, estimated prediction error, RMSE of areal prediction.

        Function do not predict unknown values, areas with NaN are skipped.

        INPUT:

        :param number_of_neighbours: (int) minimum number of neighbours to include in the algorithm,
        :param max_search_radius: (float) maximum search radius (if number of neighbours within this search radius is
            smaller than number_of_neighbours parameter then additional neighbours are included up to
            number_of_neighbours),
        :param data_crs: (str) data crs, look into: https://geopandas.org/projections.html

        OUTPUT:

        :return: regularized dataset (GeoPandas GeoDataFrame object)
        """
        areas_ids = self.areal_data_known[:, 0]

        list_of_vals = []
        for a_id in areas_ids:
            prediction_rows = self._get_prediction_row(a_id, number_of_neighbours, max_search_radius)

            # Add id and geometry into a list

            if self.ktype == 'ata':
                geometry = self.areal_data_known[self.areal_data_known[:, 0] == a_id]
                geometry = geometry[0][1]
                data_row = [a_id, geometry, prediction_rows[0], prediction_rows[1], prediction_rows[2]]
                list_of_vals.append(data_row)
            else:
                for val in prediction_rows:
                    xy = Point(val[0])
                    list_of_vals.append([a_id, xy, val[1], val[2], val[3]])

        # Transform array into a dataframe

        gdf = gpd.GeoDataFrame(list_of_vals)
        gdf.columns = ['id', 'geometry', 'estimated value', 'estimated prediction error', 'rmse']
        gdf.crs = data_crs

        return gdf



import numpy as np

from pyinterpolate.distance.calculate_distances import calc_point_to_point_distance
from pyinterpolate.transform.select_values_in_range import select_values_in_range


def calculate_covariance(data, step_size, max_range):
    """Function calculates covariance of a given set of points.

        Equation for calculation is:

            covariance = 1 / (N) * SUM(i=1, N) [z(x_i + h) * z(x_i)] - u^2

        where:

            - N - number of observation pairs,
            - h - distance (lag),
            - z(x_i) - value at location z_i,
            - (x_i + h) - location at a distance h from x_i,
            - u -mean of observations at a given lag distance.

        INPUT:

        :param data: (numpy array) coordinates and their values,
        :param step_size: (float) step size of circle within radius are analyzed points,
        :param max_range: (float) maximum range of analysis.

        OUTPUT:

        :return: covariance: numpy array of pair of lag and covariance values where:

            - covariance[0] = array of lags
            - covariance[1] = array of lag's values
            - covariance[2] = array of number of points in each lag.
    """

    distances = calc_point_to_point_distance(data[:, :-1])

    # Get only upper diagonal of distances, rest set to -1
    covar = []
    covariance = []

    lags = np.arange(0, max_range, step_size)

    for idx, h in enumerate(lags):
        distances_in_range = select_values_in_range(distances, h, step_size)
        cov = (data[distances_in_range[0], 2] * data[distances_in_range[1], 2])
        u_mean = (data[distances_in_range[0], 2] + data[distances_in_range[1], 2])
        u_mean = u_mean / (2 * len(u_mean))
        cov_value = np.sum(cov) / (len(cov)) - np.sum(u_mean) ** 2
        covar.append([cov_value, len(cov)])
        if covar[idx][0] > 0:
            covariance.append([h, covar[idx][0], covar[idx][1]])
        else:
            covariance.append([h, 0, 0])

    covariance = np.vstack(covariance)

    return covariance



import unittest
import numpy as np

from pyinterpolate.kriging.point_kriging.kriging import Krige

from pyinterpolate.semivariance.semivariogram_estimation.calculate_semivariance import calculate_semivariance
from pyinterpolate.semivariance.semivariogram_fit.fit_semivariance import TheoreticalSemivariogram
from pyinterpolate.distance.calculate_distances import calc_point_to_point_distance


class SetKrigingModel:

    def __init__(self, data, smv_model):
        self.dataset = data
        self.semivariogram_model = smv_model

        self.kriging_model = Krige(self.semivariogram_model, self.dataset)


class SetSemivariogramModel:

    def __init__(self, positions, number_of_steps):
        self.data = positions
        self.steps = number_of_steps

        distances = calc_point_to_point_distance(self.data[:, :-1])
        maximum_range = np.max(distances)
        step_size = maximum_range / self.steps

        self.semivariance = calculate_semivariance(self.data, step_size, maximum_range)
        self.theoretical_semivariance = TheoreticalSemivariogram(self.data, self.semivariance)
        self.theoretical_semivariance.find_optimal_model(number_of_ranges=self.steps)


class TestKriging(unittest.TestCase):

    @staticmethod
    def _test_ordinary_kriging(kriging_model, unknown_loc, number_of_ranges):
        km = kriging_model.ordinary_kriging(unknown_loc, number_of_ranges, test_anomalies=False)
        return int(km[0])

    @staticmethod
    def _test_simple_kriging(kriging_model, unknown_loc, number_of_ranges):
        km = kriging_model.simple_kriging(unknown_loc, number_of_ranges, test_anomalies=False)
        return int(km[0])


    def test_kriging(self):
        # Set datasets
        unknown_pos = (10, 10)
        pos = [[11, 1, 1], [23, 2, 2], [33, 3, 3], [14, 44, 4], [13, 10, 9], [12, 55, 35], [11, 9, 7]]
        pos = np.array(pos)

        # Set semivariogram and kriging model
        s_model = SetSemivariogramModel(pos, 5)
        k_model = SetKrigingModel(pos, s_model.theoretical_semivariance)

        kok = self._test_ordinary_kriging(k_model.kriging_model, unknown_pos, 6)
        ksk = self._test_simple_kriging(k_model.kriging_model, unknown_pos, 6)

        self.assertEqual(kok, 7, "Kriging value should be 7")
        self.assertEqual(ksk, kok, "Simple and Ordinary Kriging values should be the same")


if __name__ == '__main__':
    unittest.main()



from .areal_semivariance.areal_semivariance import ArealSemivariance
from .semivariogram_deconvolution.regularize_semivariogram import RegularizedSemivariogram
from .semivariogram_estimation.calculate_covariance import calculate_covariance
from .semivariogram_estimation.calculate_semivariance import calculate_semivariance, calculate_weighted_semivariance
from .semivariogram_fit.fit_semivariance import TheoreticalSemivariogram


import unittest
import os
import numpy as np
import geopandas as gpd
from pyinterpolate.kriging.areal_poisson_kriging.area_to_point.atp_poisson_kriging import AtPPoissonKriging
from pyinterpolate.semivariance.semivariogram_estimation.calculate_semivariance import calculate_weighted_semivariance
from pyinterpolate.semivariance.semivariogram_fit.fit_semivariance import TheoreticalSemivariogram

from pyinterpolate.io_ops.prepare_areal_shapefile import prepare_areal_shapefile
from pyinterpolate.io_ops.get_points_within_area import get_points_within_area
from pyinterpolate.transform.set_areal_weights import set_areal_weights

class TestAreaToPointPoissonKriging(unittest.TestCase):
    def test_atp_pk(self):
        my_dir = os.path.dirname(__file__)

        areal_dataset = os.path.join(my_dir, '../sample_data/test_areas_pyinterpolate.shp')
        subset = os.path.join(my_dir, '../sample_data/test_points_pyinterpolate.shp')

        a_id = 'id'
        areal_val = 'value'
        points_val = 'value'

        # Get maximum range and set step size

        gdf = gpd.read_file(areal_dataset)

        total_bounds = gdf.geometry.total_bounds
        total_bounds_x = np.abs(total_bounds[2] - total_bounds[0])
        total_bounds_y = np.abs(total_bounds[3] - total_bounds[1])

        max_range = min(total_bounds_x, total_bounds_y)
        step_size = max_range / 10

        areal_data_prepared = prepare_areal_shapefile(areal_dataset, a_id, areal_val)
        points_in_area = get_points_within_area(areal_dataset, subset, areal_id_col_name=a_id,
                                                points_val_col_name=points_val)

        # Get one area as unknown
        unknown_area_id = [1]

        u_points = points_in_area[points_in_area[:, 0] == unknown_area_id][0]

        k_areas = areal_data_prepared[areal_data_prepared[:, 0] != unknown_area_id]
        k_points = points_in_area[points_in_area[:, 0] != unknown_area_id]

        # Semivariance deconvolution

        semivar_modeling_data = set_areal_weights(k_areas, k_points)
        smv_model = calculate_weighted_semivariance(semivar_modeling_data, step_size, max_range)

        semivariogram = TheoreticalSemivariogram(k_areas[:, 2:], smv_model)

        semivariogram.find_optimal_model()

        # Poisson Kriging

        search_radius = max_range / 2
        number_of_observations = 3

        pkc = AtPPoissonKriging(regularized_model=semivariogram,
                                known_areas=k_areas,
                                known_areas_points=k_points)
        d = pkc.predict(u_points, number_of_observations, search_radius)

        est_areal = np.sum(d[:, 0])
        self.assertEqual(int(est_areal), 126, "Int of first value should be equal to 126")


if __name__ == '__main__':
    unittest.main()



import numpy as np
from tqdm import tqdm
from pyinterpolate.transform.select_values_in_range import select_values_in_range


def group_distances(distances_arrays, lags, step_size):
    """
    Function groups distances between blocks by the given lags and step size between lags.
    :param distances_arrays: (arrays)
        array[0] - list of distances between blocks,
        array[1] - list of blocks id;

        array[0]: [
                    distances[id A to id A, id A to id B, id A to id N],
                    distances[id B to id A, id B to id B, id B to id N],
                    distances[id N to id A, id N to id B, id N to id N]
                ],
        array[1]: [id A, id B, id N]
    :param lags: (array) lags between values,
    :param step_size: (float) step size between lags,
    :return grouped lags: (array) in the form: [lag, [[distances_in_range_for_area_1, number_of_distances],
                                                      [distances_in_range_for_area_2, number_of_distances]]]
    """
    grouped_lags = []
    for lag in lags:
        distances_list = []
        for dist in distances_arrays[0]:
            distances_in_range = select_values_in_range(dist, lag, step_size)
            distances_list.append([distances_in_range])
        grouped_lags.append([lag, distances_list])
    return grouped_lags


def calculate_average_semivariance(between_block_distances,
                                   semivariance_within_block_points,
                                   lags, step_size):
    """
    Function calculates average within-block semivariance between blocks.

    gamma_h(v, v) = 1 / (2*N(h)) SUM(from a=1 to N(h)) [gamma(va, va) + gamma(va_h, va_h)]


    :param between_block_distances: distances_arrays: (arrays)
        array[0] - list of distances between blocks,
        array[1] - list of blocks ids.

        array[0]: [
                    distances[id A to id A, id A to id B, id A to id N],
                    distances[id B to id A, id B to id B, id B to id N],
                    distances[id N to id A, id N to id B, id N to id N]
                ],
        array[1]: [id A, id B, id N]
    :param semivariance_within_block_points: (numpy array) [area_id, within block semivariance]
    :param lags: (array) lags between values,
    :param step_size: (float) step size between lags,
    :return average_semivariance: (array) [lag, average semivariance]
    """
    grouped_distances = group_distances(between_block_distances, lags, step_size)
    areas_list = between_block_distances[1]

    # Select distances per lag

    avg_semivars = []
    for distance_lag in tqdm(grouped_distances):
        avg_sem = []
        for idx, base_area_id in enumerate(areas_list):
            # Select internal semivariance of base area for chosen lag
            base_inblock_semivariance = semivariance_within_block_points[semivariance_within_block_points[:, 0] == base_area_id][0][1]
            # Check all distances in search radius
            neighbours_list = distance_lag[1][idx][0][0]
            no_of_areas = len(neighbours_list)
            # Skip if no neighbours
            if no_of_areas == 0:
                pass
            else:
                # Check if one neighbor
                # If so, check if this is the same area as base area
                # Calculate average semivariance from given area

                for neighbour in neighbours_list:
                    n_id = between_block_distances[1][neighbour]
                    n_semivar = semivariance_within_block_points[semivariance_within_block_points[:, 0] == n_id][0][1]
                    avg_sem.append((n_semivar + base_inblock_semivariance))

        if len(avg_sem) == 0:
            avg_sem = 0
        else:
            avg_sem = 0.5 * (np.mean(avg_sem))

        # Append average semivariance for given lag
        avg_semivars.append([distance_lag[0], avg_sem])

    return np.array(avg_semivars)



import numpy as np


def get_total_value_of_area(areal_points):
    total = np.sum(areal_points[1][:, 2])
    return total

def set_areal_weights(areal_data, areal_points):
    """
    Function prepares array for weighted semivariance calculation.

    INPUT:

    :param areal_data: (numpy array) of areas in the form:
        [area_id, areal_polygon, centroid coordinate x, centroid coordinate y, value],
    :param areal_points: (numpy array) of points within areas in the form:
        [area_id, [point_position_x, point_position_y, value]].

    OUTPUT:

    :return: (numpy array) of weighted points.
    """

    weighted_semivariance_input = []
    for rec in areal_data:
        rec_id = rec[0]

        # Calculate total value of points within area
        total = 0
        for points_rec in areal_points:
            if points_rec[0] == rec_id:
                total = get_total_value_of_area(points_rec)
                break
            else:
                pass

        output_record = [rec[2], rec[3], rec[4], total]
        weighted_semivariance_input.append(output_record)

    weighted_semivariance_input = np.array(weighted_semivariance_input)
    return weighted_semivariance_input



# Base libraries
import numpy as np

# Data vizualization libraries
import matplotlib.pyplot as plt

# Pyinterpolate libraries
from pyinterpolate.semivariance.areal_semivariance.areal_semivariance import ArealSemivariance
from pyinterpolate.semivariance.semivariogram_fit.fit_semivariance import TheoreticalSemivariogram
from pyinterpolate.semivariance.semivariogram_estimation.calculate_semivariance import \
    calculate_semivariance  # Experimental semivariogram


class RegularizedSemivariogram:
    """
    Class performs deconvolution of semivariogram of areal data. Whole procedure is based on the iterative process
    described in: Goovaerts P., Kriging and Semivariogram Deconvolution in the Presence of Irregular Geographical
    Units, Mathematical Geology 40(1), 101-128, 2008.

    Class works as follow:

    - initialize your object (no parameters),
    - then use fit() method to build initial point support model,
    - then use transform() method to perform semivariogram regularization,
    - save semivariogram model with export_model() method.

    Class public methods:

    fit() - fits areal data and point support data into a model, initialize experimental semivariogram,
    theoretical semivariogram model, regularized point support model and deviation.

    transform() - performs semivariogram regularization, which is an iterative process.

    export_regularized_model() - Function exports final regularized model parameters into specified csv file.

    show_baseline_semivariograms() - Function shows experimental semivariogram, initial theoretical semivariogram and
        initial regularized semivariogram after fit() operation.

    show_semivariograms() - plots experimental semivariogram of area data, theoretical curve of area data,
        regularized model values and regularized model theoretical curve.
    """

    def __init__(self):
        """
        Class has multiple params, some of them are designed to control process of regularization and other are storing
        semivariogram models and experimental (or regularized) values of semivariance.
        """

        # Procedure control parameters
        self.iter = 0
        self.max_iters = None

        self.deviation_ratio = 1
        self.min_deviation_ratio = None

        self.diff_decrease = 1
        self.min_diff_decrease = None

        self.const_d_stat_reps = 0
        self.min_diff_decrease_reps = None

        self.weight_error_lags = False

        self.weight_change = False

        self.store_models = False

        # Regularization parameters
        self.ranges = None
        self.weights = []
        self.deviations = []

        # Initial semivariogram models and parameters
        self.experimental_semivariogram_of_areal_data = None
        self.initial_theoretical_model_of_areal_data = None
        self.initial_regularized_model = None
        self.initial_deviation = None

        # Temporary semivariogram models and class parameters
        self.temp_experimental_semivariogram = None
        self.temp_theoretical_semivariogram_model = None
        self.temp_regularized_model = None
        self.temp_deviation = None

        # Optimal semivariogram models and params
        self.optimal_theoretical_model = None
        self.optimal_regularized_model = None
        self.optimal_deviation = None

        # Final models
        self.final_theoretical_model = None
        self.final_optimal_model = None

        # Data
        self.areal_data = None
        self.areal_max_range = None
        self.areal_step_size = None
        self.point_support_data = None

        # Stored models if self.store_models is True
        self.t_exp_list = []
        self.t_theo_list = []
        self.t_reg_list = []

    def _regularize(self, empirical_semivariance, semivariance_model):
        """
        Function regularizes semivariogram with ArealSemivariance class.

        INPUT:

        :param empirical_semivariance: experimental values of semivariance as an array of the form:
            [[column with lags, column with values, column with number of points within lag]],
        :param semivariance_model: TheoreticalSemivariance model,

        OUTPUT:

        :return: regularized semivariance values (array).
        """

        # Initialize areal semivariance object
        areal_semivariance = ArealSemivariance(self.areal_data,
                                               self.areal_step_size,
                                               self.areal_max_range,
                                               self.point_support_data,
                                               weighted_semivariance=self.weight_error_lags)

        # Regularize semivariogram of areal data
        theoretically_regularized_model_values = areal_semivariance.regularize_semivariogram(
            empirical_semivariance=empirical_semivariance,
            theoretical_semivariance_model=semivariance_model)

        return theoretically_regularized_model_values[:, 1]

    def _calculate_deviation(self, regularized_model, theoretical_model):
        """
        Function calculates deviation between experimental and theoretical semivariogram over given lags.

        INPUT:

        :param regularized_model: (numpy array) array of the values generated for the regularized model,
        :param theoretical_model: (TheoreticalSemivariance) theoretical model of data,

        OUTPUT:

        :return deviation: (float) scalar which describes deviation between semivariograms.
        """

        lags = self.experimental_semivariogram_of_areal_data[:, 0]
        theoretical_values = theoretical_model.predict(lags)
        regularized_values = regularized_model

        deviation = np.abs(regularized_values - theoretical_values)
        deviation = np.divide(deviation,
                              theoretical_values,
                              out=np.zeros_like(deviation),
                              where=theoretical_values != 0)
        deviation = np.mean(deviation)
        return deviation

    def _rescale_optimal_point_support(self):
        """Function rescales the optimal point support model and creates new experimental values for each lag based on
            the equation:

            y(1)(h) = y_opt(h) x w(h)

            w(h) = 1 + [(y_exp_v(h) - y_opt_v(h) / (s^2 * sqrt(iter))]

            where:

            - s = sill of the model y_exp_v
            - iter = iteration number

        OUTPUT:

        :return rescalled_point_support_semivariogram: (numpy array) of the form [[lag, semivariance, number of points]]
        """
        lags = self.experimental_semivariogram_of_areal_data[:, 0]

        y_opt_h = self.optimal_theoretical_model.predict(lags)

        if not self.weight_change:
            sill = self.initial_theoretical_model_of_areal_data.params[1]
            denominator = sill * np.sqrt(self.iter)

            y_exp_v_h = self.initial_theoretical_model_of_areal_data.predict(lags)
            y_opt_v_h = self.optimal_regularized_model

            numerator = (y_exp_v_h - y_opt_v_h)

            w = 1 + (numerator / denominator)
        else:
            w = 1 + ((self.weights[-1] - 1) / 2)

        rescalled = self.experimental_semivariogram_of_areal_data.copy()
        rescalled[:, 1] = y_opt_h * w

        return rescalled, w

    def _check_deviation_ratio(self):
        return bool(self.deviation_ratio <= self.min_deviation_ratio)

    def _check_loop_limit(self):
        return bool(self.iter >= self.max_iters)

    def _check_diff_d_stat(self):
        if self.diff_decrease < self.min_diff_decrease:

            if self.const_d_stat_reps >= self.min_diff_decrease_reps:
                return True

            self.const_d_stat_reps += 1
            return False

        if self.const_d_stat_reps >= 1:

            self.const_d_stat_reps = self.const_d_stat_reps - 1
            return False

        return False

    def _check_algorithm(self):
        t1 = self._check_deviation_ratio()  # Default False
        t2 = self._check_diff_d_stat()  # Default False
        t3 = self._check_loop_limit()  # Default False

        cond = not (t1 or t2 or t3)  # Default False

        return cond

    def fit(self, areal_data, areal_step_size, max_areal_range,
            point_support_data, ranges=16, weighted_lags=True, store_models=False):
        """
        Function fits area and point support data to the initial regularized models.

        INPUT:

        :param areal_data: (numpy array) areal data prepared with the function prepare_areal_shapefile(), where data is
            a numpy array in the form: [area_id, area_geometry, centroid x, centroid y, value],
        :param areal_step_size: (float) step size between each lag, usually it is a half of distance between lags,
        :param max_areal_range: (float) max distance to perform distance and semivariance calculations,
        :param point_support_data: (numpy array) point support data prepared with the function get_points_within_area(),
            where data is a numpy array in the form: [area_id, [point_position_x, point_position_y, value]],
        :param ranges: (int) number of ranges to test during semivariogram fitting. More steps == more accurate nugget
            and range prediction, but longer distance,
        :param weighted_lags: (bool) lags weighted by number of points; if True then during semivariogram fitting error
            of each model is weighted by number of points for each lag. In practice it means that more reliable data
            (lags) have larger weights and semivariogram is modeled to better fit to those lags,
        :param store_models: (bool) if True then experimental, regularized and theoretical models are stored in lists
            after each iteration. It is important for a debugging process.
        """

        # Update data class params
        self.areal_data = areal_data
        self.areal_max_range = max_areal_range
        self.areal_step_size = areal_step_size
        self.point_support_data = point_support_data
        self.ranges = ranges
        self.weight_error_lags = weighted_lags

        self.store_models = store_models

        # Compute experimental semivariogram of areal data from areal centroids

        areal_centroids = areal_data[:, 2:]

        self.experimental_semivariogram_of_areal_data = calculate_semivariance(
            areal_centroids,
            areal_step_size,
            max_areal_range
        )

        # Compute theoretical semivariogram of areal data from areal centroids

        self.initial_theoretical_model_of_areal_data = TheoreticalSemivariogram(
            areal_centroids,
            self.experimental_semivariogram_of_areal_data
        )

        self.initial_theoretical_model_of_areal_data.find_optimal_model(
            weighted=weighted_lags,
            number_of_ranges=self.ranges
        )

        # Regularize model
        self.initial_regularized_model = self._regularize(self.experimental_semivariogram_of_areal_data,
                                                          self.initial_theoretical_model_of_areal_data)

        # Calculate d-stat
        self.initial_deviation = self._calculate_deviation(self.initial_regularized_model,
                                                           self.initial_theoretical_model_of_areal_data)

        self.deviations.append(self.initial_deviation)

    def transform(self, max_iters=25, min_deviation_ratio=0.01, min_diff_decrease=0.01, min_diff_decrease_reps=3):
        """
        Function transofrms fitted data and performs semivariogram regularization iterative procedure.

        INPUT:

        :param max_iters: (int) maximum number of iterations,
        :param min_deviation_ratio: (float) minimum ratio between deviation and initial deviation (D(i) / D(0)) below
            each algorithm is stopped,
        :param min_diff_decrease: (float) minimum absolute difference between new and optimal deviation divided by
            optimal deviation: ABS(D(i) - D(opt)) / D(opt). If it is recorded n times (controled by
            the min_diff_d_stat_reps param) then algorithm is stopped,
        :param min_diff_decrease_reps: (int) number of iterations when algorithm is stopped if condition
            min_diff_d_stat is fulfilled.
        """

        # Check if data is fitted
        if self.initial_regularized_model is None:
            raise RuntimeError('Before transform you must fit areal data and calculate initial point support models')

        # Update class control params
        self.iter = 1
        self.max_iters = max_iters

        self.min_deviation_ratio = min_deviation_ratio

        self.min_diff_decrease = min_diff_decrease
        self.min_diff_decrease_reps = min_diff_decrease_reps

        # Update initial optimal models
        self.optimal_theoretical_model = self.initial_theoretical_model_of_areal_data
        self.optimal_regularized_model = self.initial_regularized_model
        self.optimal_deviation = self.initial_deviation

        # Prepare semivariogram modeling data
        areal_centroids = self.areal_data[:, 2:]
        ranges = self.ranges
        is_weighted = self.weight_error_lags

        # Append initial models if self.store_models is True

        if self.store_models:
            self.t_theo_list.append(self.optimal_theoretical_model)
            self.t_reg_list.append(self.optimal_regularized_model)
            self.t_exp_list.append(self.experimental_semivariogram_of_areal_data)

        # Start iteration procedure

        while self._check_algorithm():
            # Compute new experimental values for new experimental point support model

            self.temp_experimental_semivariogram, weights = self._rescale_optimal_point_support()
            self.weights.append(weights)

            # Fit rescaled empirical semivariogram to the new theoretical function
            self.temp_theoretical_semivariogram_model = TheoreticalSemivariogram(areal_centroids,
                                                                                 self.temp_experimental_semivariogram)
            self.temp_theoretical_semivariogram_model.find_optimal_model(
                weighted=is_weighted,
                number_of_ranges=ranges
            )

            # Regularize model
            self.temp_regularized_model = self._regularize(
                self.temp_experimental_semivariogram,
                self.temp_theoretical_semivariogram_model
            )

            # Compute difference statistics

            self.temp_deviation = self._calculate_deviation(self.temp_regularized_model,
                                                            self.initial_theoretical_model_of_areal_data)

            if self.temp_deviation < self.optimal_deviation:
                self.weight_change = False

                self.diff_decrease = np.abs(self.temp_deviation - self.optimal_deviation) / self.optimal_deviation
                self.deviation_ratio = self.temp_deviation / self.deviations[0]

                self.optimal_deviation = self.temp_deviation

                # Update models
                self.optimal_theoretical_model = self.temp_theoretical_semivariogram_model
                self.optimal_regularized_model = self.temp_regularized_model

            else:
                self.weight_change = True

            self.deviations.append(self.temp_deviation)
            self.iter = self.iter + 1

            # Update models if self.store_models is set to True
            if self.store_models:
                self.t_theo_list.append(self.temp_theoretical_semivariogram_model)
                self.t_exp_list.append(self.temp_experimental_semivariogram)
                self.t_reg_list.append(self.temp_regularized_model)

        # Get theoretical model from regularized
        self.final_theoretical_model = self.temp_theoretical_semivariogram_model
        self.final_optimal_model = self.optimal_regularized_model

    def export_regularized_model(self, filename):
        """
        Function exports final regularized model parameters into specified csv file.

        INPUT:

        :param filename: (str) filename for model parameters (nugget, sill, range, model type).
        """
        
        if self.final_theoretical_model is None:
            raise RuntimeError('You cannot export any model if you not transform data.')
        
        self.final_theoretical_model.export_model(filename)

    def show_baseline_semivariograms(self):
        """
        Function shows experimental semivariogram, initial theoretical semivariogram and
            initial regularized semivariogram after fit() operation.
        """
        lags = self.experimental_semivariogram_of_areal_data[:, 0]
        plt.figure(figsize=(12, 12))
        plt.plot(lags, self.experimental_semivariogram_of_areal_data[:, 1], color='b')
        plt.plot(lags, self.initial_theoretical_model_of_areal_data.predict(lags), color='r', linestyle='--')
        plt.plot(lags, self.initial_regularized_model, color='g', linestyle='-.')
        plt.legend(['Experimental semivariogram of areal data', 'Initial Semivariogram of areal data',
                    'Regularized data points'])
        plt.title('Semivariograms comparison. Deviation value: {}'.format(self.initial_deviation))
        plt.xlabel('Distance')
        plt.ylabel('Semivariance')
        plt.show()

    def show_semivariograms(self):
        """
        Function shows experimental semivariogram, theoretical semivariogram and regularized semivariogram after
            semivariogram regularization with transform() method.
        """
        lags = self.experimental_semivariogram_of_areal_data[:, 0]
        plt.figure(figsize=(12, 12))
        plt.plot(lags,
                 self.experimental_semivariogram_of_areal_data[:, 1], color='b')
        plt.plot(lags,
                 self.initial_theoretical_model_of_areal_data.predict(lags), color='r',
                 linestyle='--')
        plt.plot(lags, self.optimal_regularized_model, color='g',
                 linestyle='-.')
        plt.plot(lags,
                 self.optimal_theoretical_model.predict(lags), color='black', linestyle='dotted')
        plt.legend(['Experimental semivariogram of areal data', 'Initial Semivariogram of areal data',
                    'Regularized data points, iteration {}'.format(self.iter),
                    'Optimized theoretical point support model'])
        plt.title('Semivariograms comparison. Deviation value: {}'.format(self.optimal_deviation))
        plt.xlabel('Distance')
        plt.ylabel('Semivariance')
        plt.show()



import numpy as np


def select_values_in_range(data, lag, step_size):
    """
    Function selects set of values which are greater than (lag - step size / 2) and smaller or equal than
        (lag + step size / 2).

    INPUT:

    :param data: (numpy array) distances,
    :param lag: (float) lag within areas are included,
    :param step_size: (float) step between lags.

    OUTPUT:

    :return: numpy array mask with distances within specified radius.
    """

    step_size = step_size / 2

    # Check if numpy array is given
    if not isinstance(data, np.ndarray):
        data = np.array(data)

    greater_than = lag - step_size
    less_equal_than = lag + step_size

    # Check conditions
    condition_matrix = np.logical_and(
            np.greater(data, greater_than),
            np.less_equal(data, less_equal_than))

    # Find positions
    position_matrix = np.where(condition_matrix)
    return position_matrix



import numpy as np


def read_point_data(path, data_type):
    """
    Function reads data from a text file and converts it into numpy array.

    INPUT:

    :param path: (str) path to the file,
    :param data_type: (str) data type, available types:
        - 'txt' for txt files.

    OUTPUT:

    :return data_arr: (numpy array) of coordinates and their values."""
    if data_type == 'txt':
        data_arr = np.loadtxt(path, delimiter=',')
        return data_arr
    
    raise ValueError('Data type not supported')



import unittest
import os
import numpy as np
from pyinterpolate.io_ops.read_data import read_point_data
from pyinterpolate.distance.calculate_distances import calc_point_to_point_distance
from pyinterpolate.semivariance.semivariogram_estimation.calculate_covariance import calculate_covariance


class TestCalculateCovariance(unittest.TestCase):

    def test_calculate_covariance(self):
        my_dir = os.path.dirname(__file__)
        path = os.path.join(my_dir, '../sample_data/poland_dem_gorzow_wielkopolski')

        dataset = read_point_data(path, 'txt')

        # Set semivariance params
        distances = calc_point_to_point_distance(dataset[:, :-1])

        maximum_range = np.max(distances)
        number_of_divisions = 8
        step_size = maximum_range / number_of_divisions

        gamma = calculate_covariance(dataset, step_size, maximum_range)

        output_int = [427, 184, 0,  0,  0,  0,  0,  0]

        check_equality = (gamma[:, 1].astype(np.int) == np.array(output_int)).all()
        self.assertTrue(check_equality, "Int if output array is not equal to [427, 184, 0,  0,  0,  0,  0,  0]")


if __name__ == '__main__':
    unittest.main()



import unittest
from pyinterpolate.transform.set_areal_weights import set_areal_weights
import numpy as np


class TestArealWeightsSet(unittest.TestCase):

    def test_set_areal_weighs(self):

        rows_no = 4
        data_array_1 = np.random.random(size=(rows_no, 5))
        ids_range = np.arange(0, rows_no)
        data_array_1[:, 0] = ids_range

        da2 = []
        sums = []

        for pt in ids_range:
            x = np.random.random(size=(rows_no * 8, 3))
            row = [pt, x]
            summed = np.sum(x[:, -1])
            da2.append(row)
            sums.append(summed)

        data_array_2 = np.array(da2)

        new_set = set_areal_weights(data_array_1, data_array_2)

        for idx, s in enumerate(sums):

            # Check sums
            self.assertEqual(int(s), int(new_set[idx, -1]))

            # Check compatibility with the base ds
            self.assertTrue((data_array_1[idx, 2:] == new_set[idx, :-1]).all())


if __name__ == '__main__':
    unittest.main()



import unittest
import os
import numpy as np
from pyinterpolate.io_ops.read_data import read_point_data
from pyinterpolate.distance.calculate_distances import calc_point_to_point_distance
from pyinterpolate.semivariance.semivariogram_fit.fit_semivariance import TheoreticalSemivariogram
from pyinterpolate.semivariance.semivariogram_estimation.calculate_semivariance import calculate_semivariance
from pyinterpolate.semivariance.semivariogram_estimation.calculate_semivariance import calculate_weighted_semivariance


class TestFitSemivariance(unittest.TestCase):

    def test_fit_semivariance(self):
        my_dir = os.path.dirname(__file__)
        path = os.path.join(my_dir, '../sample_data/poland_dem_gorzow_wielkopolski')

        dataset = read_point_data(path, 'txt')
        new_col = np.arange(1, len(dataset) + 1)

        dataset_weights = np.zeros((dataset.shape[0], dataset.shape[1] + 1))
        dataset_weights[:, :-1] = dataset
        dataset_weights[:, -1] = new_col

        # Set semivariance params
        distances = calc_point_to_point_distance(dataset_weights[:, :-2])

        maximum_range = np.max(distances)
        number_of_divisions = 10
        step_size = maximum_range / number_of_divisions

        # Calculate weighted and non-weighted semivariance

        gamma_w = calculate_weighted_semivariance(dataset_weights, step_size, maximum_range)
        gamma_non = calculate_semivariance(dataset, step_size, maximum_range)

        # Fit semivariance - find optimal models
        t_non_weighted = TheoreticalSemivariogram(dataset, gamma_non)
        t_weighted = TheoreticalSemivariogram(dataset_weights[:, :-1], gamma_w)

        model_non_weighted = t_non_weighted.find_optimal_model(weighted=False, number_of_ranges=20)  # linear
        model_weighted = t_weighted.find_optimal_model(weighted=False, number_of_ranges=20)  # linear

        self.assertEqual(model_non_weighted, 'linear', "Non-weighted model should be linear")
        self.assertEqual(model_weighted, 'linear', "Weighted model should be linear")

    def test_fit_semivariance_io(self):
        # Prepare fake model for fit semivariance class

        fake_theoretical_smv = TheoreticalSemivariogram(None, None, False)

        parameters = [0, 20, 40]
        fake_theoretical_smv.params = parameters
        fmn = 'linear'
        fake_theoretical_smv.chosen_model_name = fmn

        my_dir = os.path.dirname(__file__)
        file_path = os.path.join(my_dir, '../sample_data/mock_model.csv')
        fake_theoretical_smv.export_model(file_path)

        # Clear model paramas and name

        fake_theoretical_smv.params = [None, None, None]
        fake_theoretical_smv.chosen_model_name = None

        # Check if now model is not the same
        assert fake_theoretical_smv.params != parameters
        assert fake_theoretical_smv.chosen_model_name != fmn

        # Import params
        fake_theoretical_smv.import_model(file_path)

        # Check if params are the same as at the beginning

        self.assertEqual(fake_theoretical_smv.params, parameters, "Parameters should be [0, 20, 40]")
        self.assertEqual(fake_theoretical_smv.chosen_model_name, fmn, "Model name should be linear")


if __name__ == '__main__':
    unittest.main()



import unittest
import numpy as np
from pyinterpolate.transform.prepare_kriging_data import prepare_kriging_data


class TestPrepareKrigingData(unittest.TestCase):

    def test_prepare_kriging_data(self):
        dataset_length = 3
        unknown_pos = (10, 10)
        pos = [[1, 1, 1], [2, 2, 2], [3, 3, 3], [4, 4, 4], [13, 10, 9]]
        pos = np.array(pos)
        d = prepare_kriging_data(unknown_pos, pos, dataset_length)
        vals = np.array([[13, 10, 9, 3], [4, 4, 4, 8], [3, 3, 3, 9]])
        d = d.astype(np.int)

        test_output = (d == vals).all()

        test_length = len(d) == dataset_length

        self.assertTrue(test_length, "Length of prepared dataset should be 3")
        self.assertTrue(test_output, "Output array should be [[13, 10, 9, 3], [4, 4, 4, 8], [3, 3, 3, 9]]")


if __name__ == '__main__':
    unittest.main()



import os
import sample_data


class SampleData:

    def __init__(self):
        self.module_path = os.path.dirname(sample_data.__file__)
        self.dem = os.path.join(self.module_path, 'poland_dem_gorzow_wielkopolski')
        self.accidents = os.path.join(self.module_path, 'areal_data/road_accidents.shp')
        self.population = os.path.join(self.module_path, 'population_data/population_centroids_poland.shp')



import numpy as np

from pyinterpolate.transform.prepare_kriging_data import prepare_ata_known_areas
from pyinterpolate.transform.prepare_kriging_data import prepare_atp_data


class WeightedBlock2BlockSemivariance:

    def __init__(self, semivariance_model):
        self.semivariance_model = semivariance_model

    def _avg_smv(self, datarows):
        """
        Function calculates weight and partial semivariance of a data row [point value a, point value b, distance]
        :param datarows: (numpy arrays) numpy arrays of [point value a, point value b, distance between points]
        :return: (float, float) weighted semivariance and single weight
        """
        weighted_semivariance_sum = 0
        sum_of_weights = 0

        for datarow in datarows:
            single_weight = datarow[0] * datarow[1]
            partial_semivar = self.semivariance_model.predict(datarow[-1])
            weighted_semivar = partial_semivar * single_weight
            weighted_semivariance_sum = weighted_semivariance_sum + weighted_semivar
            sum_of_weights = sum_of_weights + single_weight

        return weighted_semivariance_sum, sum_of_weights

    def calculate_average_semivariance(self, data_points):
        """
        Function calculates average semivariance. Calculation is performed based on the equation:

            (1 / SUM(Pi-s=1) SUM(Pj-s'=1) w_(ss')) * SUM(Pi-s=1) SUM(Pj-s'=1) w_(ss') * semivariance(distance)

        :return: (numpy array) array of predicted and weighted semivariances
        """
        k = []
        for prediction_input in data_points:
            w_sem_sum = 0
            w_sem_weights_sum = 0
            for single_area in prediction_input:
                w_sem = self._avg_smv(single_area)
                w_sem_sum = w_sem_sum + w_sem[0]
                w_sem_weights_sum = w_sem_weights_sum + w_sem[1]
            k.append(w_sem_sum / w_sem_weights_sum)
        return np.array(k)


class WeightedBlock2PointSemivariance:

    def __init__(self, semivariance_model):
        self.semivariance_model = semivariance_model

    def _avg_smv(self, unknown_pt_value, known_points_values_and_distances):
        """
        Function calculates weight and partial semivariance of a data row [point value a, point value b, distance]
        :param unknown_pt_value: (float) value of an unknown point,
        :param known_points_values_and_distances: (numpy array) numpy array of known points values and their distances
                                                                to the unknown point,
        :return: (float) weighted semivariance for a given point
        """

        if unknown_pt_value > 0:
            partial_semivars = np.array(
                [self.semivariance_model.predict(x) for x in known_points_values_and_distances[:, 1]]
            )

            all_weights = unknown_pt_value * known_points_values_and_distances[:, 0]

            weighted_semivars = np.sum(partial_semivars * all_weights)
            weighted_block_smv = weighted_semivars / np.sum(all_weights)
        else:
            weighted_block_smv = 0

        return weighted_block_smv

    def calculate_average_semivariance(self, data_points):
        """
        Function calculates average semivariance between block (Pi) and single point from a different block Pj.
        Calculation is performed based on the equation:

            (1 / SUM(Pi-s=1) (Pj-s'=1) w_(ss')) * SUM(Pi-s=1) (Pj-s'=1) w_(ss') * semivariance(distance)

        :param data_points: (numpy array) values of points of unknown location and arrays of values of known points
                                          and distances to these points from an uknown point,
        :return: (numpy array) array of predicted and weighted semivariances for each passed point
        """

        point_to_blocks_smvs = []
        for known_area in data_points:
            pts_per_area = []
            for single_unknown_point in known_area:
                pt_output = self._avg_smv(single_unknown_point[0], single_unknown_point[1])
                pts_per_area.append(pt_output)
            point_to_blocks_smvs.append(pts_per_area)
        return np.array(point_to_blocks_smvs)


class AtPPoissonKriging:

    def __init__(self, regularized_model=None, known_areas=None, known_areas_points=None):
        """

        :param regularized_model: (RegularizedSemivariogram object) Semivariogram model after deconvolution procedure,
        :param known_areas: (numpy array) array of areas in the form:
            [area_id, areal_polygon, centroid coordinate x, centroid coordinate y, value]
        :param known_areas_points: (numpy array) array of points within areas in the form:
            [area_id, [point_position_x, point_position_y, value]]
        """

        self.model = regularized_model
        self.known_areas = known_areas
        self.known_areas = self.known_areas[self.known_areas[:, 0].argsort()]  # Sort by id
        self.known_areas_points = known_areas_points
        self.known_areas_points = self.known_areas_points[self.known_areas_points[:, 0].argsort()]  # Sort by id
        self.unknown_points = None
        self.unknown_area_total_population = None

        self.block_to_block_smv = None
        self.block_to_point_smv = None
        self.prepared_data = None

    @staticmethod
    def _add_one_to_vecs(vecs_list):
        ones = np.ones(np.shape(vecs_list)[1])
        list_with_ones = np.vstack((vecs_list, ones))
        return list_with_ones

    def predict(self, unknown_location_points,
                number_of_neighbours, max_search_radius):
        """
        Function predicts areal value in a unknown location based on the centroid-based Poisson Kriging
        :param unknown_location_points: (numpy array) array of points within an unknown area in the form:
            [area_id, [point_position_x, point_position_y, value]]
        :param number_of_neighbours: (int) minimum number of neighbours to include in the algorithm,
        :param max_search_radius: (float) maximum search radius (if number of neighbours within this search radius is
            smaller than number_of_neighbours parameter then additional neighbours are included up to number of
            neighbors).
        :return: prediction, error, estimated mean, weights:
            [value in unknown location, error, estimated mean, weights, point coordinates]
        """

        # Get total population of the unknown area
        self.unknown_area_total_population = np.sum(unknown_location_points[1][:, -1])

        self.prepared_data, self.unknown_points = prepare_atp_data(
            points_within_unknown_area=unknown_location_points,
            known_areas=self.known_areas, points_within_known_areas=self.known_areas_points,
            number_of_neighbours=number_of_neighbours, max_search_radius=max_search_radius
        )  # [[id (known), val, [
           #                    [unknown_point_value
           #                        [known pt vals, distances between unknown point and known points]],
           #                    [...],
           #                   ],
           #  total points value], [list of unknown point coordinates]]

        self.block_to_point_smv = WeightedBlock2PointSemivariance(semivariance_model=self.model)
        self.block_to_block_smv = WeightedBlock2BlockSemivariance(semivariance_model=self.model)

        # Calculate Average Point to Area semivariance for each area - len of dataset
        u_pts_and_areas = self.prepared_data[:, 2]
        k = self.block_to_point_smv.calculate_average_semivariance(u_pts_and_areas)

        # ADD ONE TO THE EACH VECTOR

        nk = self._add_one_to_vecs(k)

        # Calculate block to block average semivariance between known areas

        # Prepare areas blocks for calculation

        block_ids = self.prepared_data[:, 0]

        points_of_interest = []
        for area_id in block_ids:
            points_of_interest.append(self.known_areas_points[self.known_areas_points[:, 0] == area_id])

        # Calculate distances between blocks / prepare data for prediction
        # [id base,
        #     [
        #         id other,
        #         [base point value, other point value, distance between points]
        #     ]
        #  ]

        distances_of_chosen_areas = prepare_ata_known_areas(points_of_interest)

        predicted = self._calculate_avg_semivars_between_known(distances_of_chosen_areas)

        # Add weights to predicted values (diagonal)

        weights_mtx = self.calculate_weight_arr()
        predicted = predicted + weights_mtx

        # Prepare weights matrix

        p_ones = np.ones((predicted.shape[0], 1))
        predicted_with_ones_col = np.c_[predicted, p_ones]
        p_ones_row = np.ones((1, predicted_with_ones_col.shape[1]))
        p_ones_row[0][-1] = 0.
        weights = np.r_[predicted_with_ones_col, p_ones_row]

        # Solve Kriging systems
        # Each point of unknown area represents different kriging system

        # Transpose points matrix
        transposed = nk.T

        predicted = []

        for idx, point in enumerate(transposed):

            analyzed_pts = unknown_location_points[1][idx, :-1]
            weights = weights.astype(np.float)
            point = point.astype(np.float)

            w = np.linalg.solve(weights, point)

            zhat = self.prepared_data[:, 1].dot(w[:-1])

            if (zhat < 0) or (zhat == np.nan):
                zhat = 0

            point_pop = unknown_location_points[1][idx, -1]
            zhat = (zhat * point_pop) / self.unknown_area_total_population

            # Calculate error

            sigmasq = (w.T * point)[0]
            if sigmasq < 0:
                sigma = 0
            else:
                sigma = np.sqrt(sigmasq)
            predicted.append([zhat, sigma, w[-1], w, analyzed_pts])

        return np.array(predicted)  # return list of lists: [zhat, sigma, mean, weights, point coordinates]

    def calculate_weight_arr(self):
        """Function calculates additional weights for the diagonal matrix of predicted values"""

        vals_of_neigh_areas = self.prepared_data[:, 1]
        pop_of_neigh_areas = self.prepared_data[:, -1]

        weighted = np.sum(vals_of_neigh_areas * pop_of_neigh_areas)
        weights_arr = weighted / np.sum(pop_of_neigh_areas)
        w = np.ones(shape=vals_of_neigh_areas.shape)
        w = (weights_arr * w) / pop_of_neigh_areas
        return np.diag(w)

    def _calculate_avg_semivars_between_known(self, list_of_distances):
        """
        Function calculates average semivariance between known areas
        :param list_of_distances: (numpy array)
            [id base,
                [
                    id other,
                    [base point value, other point value, distance between points]
                ]
            ]
        :return: (numpy array) array of predicted and weighted semivariances
        """

        semivariances = []
        for area in list_of_distances:
            points_group = area[1]
            points_group = [x[1] for x in points_group]
            smv = self.block_to_block_smv.calculate_average_semivariance(points_group)
            semivariances.append(smv)
        return np.array(semivariances)



import numpy as np
from pyinterpolate.distance.calculate_distances import calc_point_to_point_distance,\
    calc_block_to_block_distance
from pyinterpolate.transform.set_areal_weights import get_total_value_of_area


def prepare_kriging_data(unknown_position, data_array, number_of_neighbours=10):
    """
    Function prepares data for kriging - array of point position, value and distance to an unknown point.

    INPUT:

    :param unknown_position: (numpy array) position of unknown value,
    :param data_array: (numpy array) known positions and their values,
    :param number_of_neighbours: (int) number of the closest locations to the unknown position.

    OUTPUT:
    :return: (numpy array) dataset with position, value and distance to the unknown point:
        [[x, y, value, distance to unknown position], [...]]
    """

    # Distances to unknown point
    r = np.array([unknown_position])

    known_pos = data_array[:, :-1]
    dists = calc_point_to_point_distance(r, known_pos)

    # Prepare data for kriging
    kriging_output_array = np.c_[data_array, dists.T]
    kriging_output_array = kriging_output_array[kriging_output_array[:, -1].argsort()]
    prepared_data = kriging_output_array[:number_of_neighbours]

    return prepared_data


def prepare_poisson_kriging_data(unknown_area, points_within_unknown_area,
                                 known_areas, points_within_known_areas,
                                 number_of_neighbours, max_search_radius,
                                 weighted=False):
    """
    Function prepares data for centroid based Poisson Kriging.

    INPUT:

    :param unknown_area: (numpy array) unknown area in the form:
        [area_id, polygon, centroid x, centroid y],
    :param points_within_unknown_area: (numpy array) points and their values within the given area:
        [area_id, [point_position_x, point_position_y, value]],
    :param known_areas: (numpy array) known areas in the form:
        [area_id, polygon, centroid x, centroid y, aggregated value],
    :param points_within_known_areas: (numpy array) points and their values within the given area:
        [area_id, [point_position_x, point_position_y, value]],
    :param number_of_neighbours: (int) minimum number of neighbours to include in the algorithm,
    :param max_search_radius: (float) maximum search radius (if number of neighbours within this search radius is
        smaller than number_of_neighbours parameter then additional neighbours are included up to number of neighbors),
    :param weighted: (bool) distances weighted by population (True) or not (False).

    OUTPUT:

    :return: (numpy array) distances from known locations to the unknown location:
        [id_known, coordinate x, coordinate y, value, distance to unknown, aggregated points values within an area].
    """

    # Prepare data
    cx_cy = unknown_area[2:-1]
    r = np.array(cx_cy)

    known_centroids = known_areas.copy()
    kc_ids = known_centroids[:, 0]
    kc_vals = known_centroids[:, -1]
    kc_pos = known_centroids[:, 2:-1]

    # Build set for Poisson Kriging

    if weighted:
        known_areas_pts = points_within_known_areas.copy()

        dists = []  # [id_known, dist]

        for pt in known_areas_pts:
            d = calc_block_to_block_distance([pt, points_within_unknown_area])
            dists.append([d[0][0][1]])
        s = np.ravel(np.array(dists)).T
        kriging_data = np.c_[kc_ids, kc_pos, kc_vals, s]  # [id, coo_x, coo_y, val, dist_to_unkn]
    else:
        dists = calc_point_to_point_distance(kc_pos, [r])
        dists = dists.ravel()
        s = dists.T
        kriging_data = np.c_[kc_ids, kc_pos, kc_vals, s]  # [id, coo_x, coo_y, val, dist_to_unkn]

    # sort by distance
    kriging_data = kriging_data[kriging_data[:, -1].argsort()]

    # Get distances in max search radius
    max_search_pos = np.argmax(kriging_data[:, -1] > max_search_radius)
    output_data = kriging_data[:max_search_pos]

    # check number of observations

    if len(output_data) < number_of_neighbours:
        output_data = kriging_data[:number_of_neighbours]

    # get total points' value in each id from prepared datasets and append it to the array

    points_vals = []
    for rec in output_data:
        areal_id = rec[0]
        points_in_area = points_within_known_areas[points_within_known_areas[:, 0] == areal_id]
        total_val = get_total_value_of_area(points_in_area[0])
        points_vals.append(total_val)

    output_data = np.c_[output_data, np.array(points_vals)]
    return output_data


def _merge_vals_and_distances(known_vals, unknown_vals, distances_array):
    """
    Function prepares array of point values and respective distances for Poisson Kriging distance
    :param known_vals: (numpy array) list of known area point values - number of rows of output array,
    :param unknown_vals: (numpy array) list of unknown area point values - number of columns of output array,
    :param distances_array: (numpy array) distances array with the same number of rows as known_vals and
        the same number of columns as unknown_vals arrays,
    :return output_arr: (numpy array) array of [known point value, unknown point value, distance between points]
    """
    output = []
    for k_idx, value in enumerate(known_vals):
        output_list = [[value, x, distances_array[k_idx, u_idx]] for u_idx, x in enumerate(unknown_vals)]
        output.append(output_list)
    output_arr = np.array(output)
    return output_arr


def prepare_ata_data(points_within_unknown_area,
                     known_areas, points_within_known_areas,
                     number_of_neighbours, max_search_radius):
    """
    Function prepares data for Area to Area Poisson Kriging.

    INPUT:

    :param points_within_unknown_area: (numpy array) points and their values within the unknown area:
        [area_id, [point_position_x, point_position_y, value of point]],
    :param known_areas: (numpy array) known areas in the form:
        [area_id, areal_polygon, centroid coordinate x, centroid coordinate y, value at specific location],
    :param points_within_known_areas: (numpy array) points and their values within the given area:
        [[area_id, [point_position_x, point_position_y, value of point]], ...],
    :param number_of_neighbours: (int) minimum number of neighbours to include in the algorithm,
    :param max_search_radius: (float) maximum search radius (if number of neighbours within this search radius is
        smaller than number_of_neighbours parameter then additional neighbours are included up to number of neighbors).

    OUTPUT:

    :return output_data: (numpy array) distances from known locations to the unknown location: [id (known),
        areal value - count, [known_point_1 value, unknown_point_1 value, distance_1], total point value].
    """

    # Initialize set

    kriging_areas_ids = known_areas[:, 0]
    kriging_areal_values = known_areas[:, -1]

    # Build set for Area to Area Poisson Kriging - sort areas with distance

    known_areas_pts = points_within_known_areas.copy()

    dists = []  # [id_known, dist to unknown]

    for pt in known_areas_pts:
        d = calc_block_to_block_distance([pt, points_within_unknown_area])
        dists.append([d[0][0][1]])
    s = np.ravel(np.array(dists)).T
    kriging_data = np.c_[kriging_areas_ids, kriging_areal_values, s]  # [id, areal val, dist_to_unkn]

    # sort by distance
    kriging_data = kriging_data[kriging_data[:, -1].argsort()]

    # Get distances in max search radius
    max_search_pos = np.argmax(kriging_data[:, -1] > max_search_radius)
    output_data = kriging_data[:max_search_pos]

    # check number of observations

    if len(output_data) < number_of_neighbours:
        output_data = kriging_data[:number_of_neighbours]

    # for each of prepared id prepare distances list with points' weights for semivariogram calculation

    points_vals = []
    points_in_unknown_area = points_within_unknown_area[1][:, :-1]
    vals_in_unknown_area = points_within_unknown_area[1][:, -1]
    for rec in output_data:
        areal_id = rec[0]
        areal_value = rec[1]
        known_area = points_within_known_areas[points_within_known_areas[:, 0] == areal_id]
        known_area = known_area[0]
        points_in_known_area = known_area[1][:, :-1]
        vals_in_known_area = known_area[1][:, -1]
        distances_array = calc_point_to_point_distance(points_in_known_area, points_in_unknown_area)
        merged = _merge_vals_and_distances(vals_in_known_area, vals_in_unknown_area, distances_array)
        total_val = np.sum(known_area[1][:, 2])
        generated_array = [areal_id, areal_value, merged, total_val]  # [id, value, [known point value,
                                                                      #              unknown point value,
                                                                      #              distance between points],
                                                                      #              total point value]
        points_vals.append(generated_array)

    output_data = np.array(points_vals)
    return output_data


def prepare_ata_known_areas(list_of_points_of_known_areas):
    """
    Function prepares known areas data for prediction.

    INPUT:

    :param list_of_points_of_known_areas: (numpy array) list of all areas' points and their values used for the
        prediction.

    OUTPUT:

    :return: (numpy array) list of arrays with areas and distances between them:
        [id base, [id other, [base point value, other point value,  distance between points]]].
    """
    all_distances_list = []
    for pt1 in list_of_points_of_known_areas:

        id_base = pt1[0][0]
        list_of_distances_from_base = [id_base, []]

        points_in_base_area = pt1[0][1][:, :-1]
        vals_in_base_area = pt1[0][1][:, -1]

        for pt2 in list_of_points_of_known_areas:

            id_other = pt2[0][0]
            points_in_other_area = pt2[0][1][:, :-1]
            vals_in_other_area = pt2[0][1][:, -1]

            distances_array = calc_point_to_point_distance(points_in_base_area, points_in_other_area)
            merged = _merge_vals_and_distances(vals_in_base_area, vals_in_other_area, distances_array)

            list_of_distances_from_base[1].append([id_other, merged])
        all_distances_list.append(list_of_distances_from_base)

    return np.array(all_distances_list)


def prepare_distances_list_unknown_area(unknown_area_points):
    """
    Function prepares distances list of unknown (single) area.

    INPUT:

    :param unknown_area_points: [pt x, pt y, val].

    OUTPUT:

    :return: [point value 1, point value 2,  distance between points].
    """
    dists = calc_point_to_point_distance(unknown_area_points[:, :-1])
    vals = unknown_area_points[:, -1]

    merged = _merge_vals_and_distances(vals, vals, dists)
    return np.array(merged)


def _merge_point_val_and_distances(unknown_point_val, known_vals, distances_array):
    """
    Function prepares array of point values and respective distances for Poisson Kriging distance.

    INPUT:

    :param unknown_point_val: (float) unknown point value,
    :param known_vals: (numpy array) list of unknown area point values,
    :param distances_array: (numpy array) distances from unknown area point to known area points.

    OUTPUT:

    :return output_arr: (numpy array) array of [unknown point value, [known points values, distance between points]].
    """
    distances_array = distances_array[:, 0]
    otp = np.array([list(x) for x in zip(known_vals, distances_array)])

    output_arr = np.array([unknown_point_val, otp])
    return output_arr

def prepare_atp_data(points_within_unknown_area,
                     known_areas, points_within_known_areas,
                     number_of_neighbours, max_search_radius):
    """
    Function prepares data for Area to Point Poisson Kriging.

    INPUT:

    :param points_within_unknown_area: (numpy array) points and their values within the given area:
        [area_id, [point_position_x, point_position_y, value of point]],
    :param known_areas: (numpy array) known areas in the form:
        [area_id, areal_polygon, centroid coordinate x, centroid coordinate y, value at specific location],
    :param points_within_known_areas: (numpy array) points and their values within the given area:
        [[area_id, [point_position_x, point_position_y, value of point]], ...],
    :param number_of_neighbours: (int) minimum number of neighbours to include in the algorithm,
    :param max_search_radius: (float) maximum search radius (if number of neighbours within this search radius is
        smaller than number_of_neighbours parameter then additional neighbours are included up to number of neighbors).

    OUTPUT:

    :return output_data: (numpy array) distances from known locations to the unknown location:
        [
            id_known,
            areal value - count,
            [
                unknown point value,
                [known point values, distance],
            total point value count
            ],
            [array of unknown area points coordinates]
        ]
    """

    # Initialize set

    kriging_areas_ids = known_areas[:, 0]
    kriging_areal_values = known_areas[:, -1]

    # Build set for Area to Area Poisson Kriging - sort areas with distance

    known_areas_pts = points_within_known_areas.copy()

    dists = []  # [id_known, dist to unknown]

    for pt in known_areas_pts:
        d = calc_block_to_block_distance([pt, points_within_unknown_area])
        dists.append([d[0][0][1]])
    s = np.ravel(np.array(dists)).T
    kriging_data = np.c_[kriging_areas_ids, kriging_areal_values, s]  # [id, areal val, dist_to_unkn]

    # sort by distance
    kriging_data = kriging_data[kriging_data[:, -1].argsort()]

    # Get distances in max search radius
    max_search_pos = np.argmax(kriging_data[:, -1] > max_search_radius)
    output_data = kriging_data[:max_search_pos]

    # check number of observations

    if len(output_data) < number_of_neighbours:
        output_data = kriging_data[:number_of_neighbours]

    # for each of prepared id prepare distances list with points' weights for semivariogram calculation

    points_vals = []
    points_and_vals_in_unknown_area = points_within_unknown_area[1]
    for rec in output_data:
        areal_id = rec[0]
        areal_value = rec[1]
        known_area = points_within_known_areas[points_within_known_areas[:, 0] == areal_id]
        known_area = known_area[0]
        points_in_known_area = known_area[1][:, :-1]
        vals_in_known_area = known_area[1][:, -1]

        # Set distances array from each point of unknown area
        merged_points_array = []
        for u_point in points_and_vals_in_unknown_area:
            u_point_dists = calc_point_to_point_distance(points_in_known_area, [u_point[:-1]])
            u_point_val = u_point[-1]
            merged = _merge_point_val_and_distances(u_point_val, vals_in_known_area, u_point_dists)
            merged_points_array.append(merged)

        total_val = np.sum(known_area[1][:, 2])
        generated_array = [areal_id, areal_value, merged_points_array, total_val]  # [[id, value, [
                                                                                   # [unknown point value,
                                                                                   #     [known points values,
                                                                                   #      distances between points]],
                                                                                   # ...],
                                                                                   #  total known points value],
                                                                                   # [list of uknown point coords]]
        points_vals.append(generated_array)

    output_data = np.array(points_vals)
    return [output_data, points_within_unknown_area[1][:, :-1]]



import csv
from operator import itemgetter

import numpy as np
import matplotlib.pyplot as plt


class TheoreticalSemivariogram:
    """
    Class calculates theoretical semivariogram. Class takes two parameters during initialization:

        points_array - (numpy array) analysed points where the last column is representing values, typically x, y,
            value,
        empirical_semivariance - (numpy array) semivariance where first row of array represents lags and the second row
            represents semivariance's values for given lag.

    Available methods:

    - predict() - method predicts value of the unknown point based on the chosen model,
    - fit_semivariance() - Method fits experimental points into chosen semivariance model type,
    - find_optimal_model() - Method fits experimental points into all available models and choose one with the lowest
        error.

    Available theoretical models:

    - spherical_model(distance, nugget, sill, semivar_range)
    - gaussian_model(distance, nugget, sill, semivar_range)
    - exponential_model(distance, nugget, sill, semivar_range)
    - linear_model(distance, nugget, sill, semivar_range)

    Additional methods:

    - calculate_base_error(),
    - show_experimental_semivariogram() - shows semivariogram which is a part of the class object's instance,
    - show_semivariogram() - shows experimental semivariogram with theoretical model (if it was calculated).
    """

    def __init__(self, points_array=None, empirical_semivariance=None, verbose=False):
        """
        INPUT:

        :param points_array: (numpy array) [point x, point y, value] (optional if model parameters are imported)
        :param empirical_semivariance: (numpy array) array of pair of lag and semivariance values where
                 semivariance[:, 0] = array of lags
                 semivariance[:, 1] = array of lag's values
                 semivariance[:, 2] = array of number of points in each lag.
                 (optional if model parameters are imported)
        :param verbose: (bool) if True then all messages are printed, otherwise nothing.
        """

        self.points_values = points_array
        self.empirical_semivariance = empirical_semivariance
        self.verbose = verbose
        self.theoretical_model = None
        self.chosen_model_name = None
        self.params = None
        self.model_error = None

    # MODELS

    @staticmethod
    def spherical_model(distance, nugget, sill, semivar_range):
        """

        INPUT:

        :param distance: array of ranges from empirical semivariance,
        :param nugget: scalar,
        :param sill: scalar,
        :param semivar_range: optimal range calculated by fit_semivariance method.

        OUTPUT:

        :return: an array of modeled values for given range. Values are calculated based on the spherical model.
        """

        x = np.where((distance <= semivar_range),
                     (nugget + sill * (3.0 * distance / (2.0 * semivar_range)) - (
                             (distance / semivar_range) ** 3.0 / 2.0)),
                     (nugget + sill))
        return x

    @staticmethod
    def exponential_model(distance, nugget, sill, semivar_range):
        """

        INPUT:

        :param distance: array of ranges from empirical semivariance,
        :param nugget: scalar,
        :param sill: scalar,
        :param semivar_range: optimal range calculated by fit_semivariance method.

        OUTPUT:

        :return: an array of modeled values for given range. Values are calculated based on the exponential model.
        """
        try:
            x = nugget + sill * (1 - np.exp(-distance / semivar_range))
        except TypeError:
            distance = distance.astype(float)
            semivar_range = float(semivar_range)
            x = nugget + sill * (1 - np.exp(-distance / semivar_range))
        return x

    @staticmethod
    def linear_model(distance, nugget, sill, semivar_range):
        """

        INPUT:

        :param distance: array of ranges from empirical semivariance,
        :param nugget: scalar,
        :param sill: scalar,
        :param semivar_range: optimal range calculated by fit_semivariance method.

        OUTPUT:

        :return: an array of modeled values for given range. Values are calculated based on the linear model.
        """

        x = np.where((distance <= semivar_range),
                     (nugget + sill * (distance / semivar_range)),
                     (nugget + sill))
        return x

    @staticmethod
    def gaussian_model(distance, nugget, sill, semivar_range):
        """

        INPUT:

        :param distance: array of ranges from empirical semivariance,
        :param nugget: scalar,
        :param sill: scalar,
        :param semivar_range: optimal range calculated by fit_semivariance method.

        OUTPUT:

        :return: an array of modeled values for given range. Values are calculated based on the gaussian model.
        """
        x = nugget + sill * (1 - np.exp(-distance * distance / (semivar_range ** 2)))
        return x

    def fit_semivariance(self, model_type, number_of_ranges=16):
        """
        Method fits experimental points into chosen semivariance model type.

        INPUT:

        :param model_type: (str) 'exponential', 'gaussian', 'linear', 'spherical',
        :param number_of_ranges: (int) deafult = 16. Used to create an array of equidistant ranges
            between minimal range of empirical semivariance and maximum range of empirical semivariance.

        OUTPUT:

        :return: (model_type, model parameters)
        """

        # model
        models = {
            'spherical': self.spherical_model,
            'exponential': self.exponential_model,
            'linear': self.linear_model,
            'gaussian': self.gaussian_model,
        }
        model = models[model_type]
        self.chosen_model_name = model_type
        self.theoretical_model = model

        # sill
        sill = np.var(self.points_values[:, -1])

        # nugget
        if self.empirical_semivariance[0][0] != 0:
            nugget = 0
        else:
            nugget = self.empirical_semivariance[0][1]

        # range
        minrange = self.empirical_semivariance[:, 0][1]
        maxrange = self.empirical_semivariance[:, 0][-1]
        ranges = np.linspace(minrange, maxrange, number_of_ranges)
        optimal_range = self.calculate_range(model, ranges, nugget, sill)

        # output model
        self.params = [nugget, sill, optimal_range]

        # model error
        self.model_error = self.calculate_model_error(model, self.params)

        return model_type, self.params

    def find_optimal_model(self, weighted=False, number_of_ranges=16):
        """

        Method fits experimental points into all available models and choose one with the lowest error.

        INPUT:

        :param weighted: (bool) default=False. If True then each lag is weighted by:

            sqrt(N(h))/gamma_{exp}(h)

            where:

            - N(h) - number of point pairs in a given range, gamma_{exp}(h) - value of experimental semivariogram for h.

        :param number_of_ranges: (int) default=16. Used to create an array of equidistant ranges
            between minimal range of empirical semivariance and maximum range of empirical semivariance.
        """

        # models
        models = {
            'spherical': self.spherical_model,
            'exponential': self.exponential_model,
            'linear': self.linear_model,
        }

        # calculate base error for a flat line
        base_error = self.calculate_base_error()

        # sill
        sill = np.var(self.points_values[:, -1])

        # nugget
        if self.empirical_semivariance[0][0] != 0:
            nugget = 0
        else:
            nugget = self.empirical_semivariance[0][1]

        # range
        minrange = self.empirical_semivariance[:, 0][1]
        maxrange = self.empirical_semivariance[:, 0][-1]
        ranges = np.linspace(minrange, maxrange, number_of_ranges)

        # Calculate model errors
        model_errors = [('NULL model', base_error, None)]
        for model in models:
            optimal_range = self.calculate_range(models[model], ranges, nugget, sill)

            # output model
            params = [nugget, sill, optimal_range]
            if not weighted:
                model_error = self.calculate_model_error(models[model], params)
            else:
                model_error = self.calculate_model_error(models[model], params, True)

            model_errors.append((model, model_error, params))
            if self.verbose:
                print('Model: {}, error value: {}'.format(model, model_error))

        # Select the best model
        sorted_errors = sorted(model_errors, key=itemgetter(1))

        if sorted_errors[0][0] == 'NULL model':
            # This is unlikely case when error estimated as the squared distance between extrapolated values and
            # x axis is smaller than models' errors

            model_name = sorted_errors[1][0]
            model_error = sorted_errors[1][1]
            model_params = sorted_errors[1][2]
            self.theoretical_model = models[model_name]
            self.params = model_params

            print('WARNING: NULL model error is better than estimated models, its value is:', sorted_errors[0][1])
            print('Chosen model: {}, with value of: {}.'.format(
                model_name, model_error
            ))

        else:

            model_name = sorted_errors[0][0]
            model_error = sorted_errors[0][1]
            model_params = sorted_errors[0][2]
            self.theoretical_model = models[model_name]
            self.params = model_params
            if self.verbose:
                print('Chosen model: {}, with value: {}.'.format(
                    model_name, model_error
                ))
        self.chosen_model_name = model_name
        self.model_error = model_error
        return model_name

    def calculate_range(self, model, ranges, nugget, sill):
        errors = []
        for r in ranges:
            x = (self.empirical_semivariance[:, 1] - model(self.empirical_semivariance[:, 0], nugget, sill, r))
            x = x ** 2
            errors.append(np.mean(x))
        range_pos = np.argmin(errors)
        optimal_rg = ranges[range_pos]
        return optimal_rg

    def calculate_values(self):
        output_model = self.theoretical_model(self.empirical_semivariance[:, 0],
                                              self.params[0],
                                              self.params[1],
                                              self.params[2])
        return output_model

    def calculate_base_error(self):
        """
        Method calculates base error as a squared difference between experimental semivariogram and
        a "flat line" on the x-axis (only zeros)
        """
        n = len(self.empirical_semivariance[:, 1])
        zeros = np.zeros(n)
        error = np.mean(np.abs(self.empirical_semivariance[:, 1] - zeros))
        return error

    def calculate_model_error(self, model, par, weight=False):
        if not weight:
            error = np.abs(self.empirical_semivariance[:, 1] - model(self.empirical_semivariance[:, 0],
                                                                     par[0],
                                                                     par[1],
                                                                     par[2]))
        else:
            nh = np.sqrt(self.empirical_semivariance[:, 2])
            vals = self.empirical_semivariance[:, 1]
            nh_divided_by_vals = np.divide(nh,
                                           vals,
                                           out=np.zeros_like(nh),
                                           where=vals != 0)
            error = np.abs(nh_divided_by_vals *
                           (vals - model(self.empirical_semivariance[:, 0], par[0], par[1], par[2])))
        error = np.mean(error)
        return error

    def predict(self, distances):
        """

        INPUT:

        :param distances: array of distances from points of known locations and values to the point of
        unknown value,

        OUTPUT:
        :return: model with predicted values.
        """

        output_model = self.theoretical_model(distances,
                                              self.params[0],
                                              self.params[1],
                                              self.params[2])
        return output_model

    def export_model(self, filename):
        """
        Function exports semivariance model to the csv file with columns:

        - name: [model name],
        - nugget: [value],
        - sill: [value],
        - range: [value]"""

        model_parameters = {
            'name': self.chosen_model_name,
            'nugget': self.params[0],
            'sill': self.params[1],
            'range': self.params[2],
        }

        csv_cols = list(model_parameters.keys())
        try:
            with open(filename, 'w') as semivar_csv:
                writer = csv.DictWriter(semivar_csv, fieldnames=csv_cols)
                writer.writeheader()
                writer.writerow(model_parameters)
        except IOError:
            print("I/O error, provided path is not valid")

    def import_model(self, filename):
        """

        Function imports semivariance model and updates it's parameters (model name, nugget, sill, range)."""

        models = {
            'spherical': self.spherical_model,
            'exponential': self.exponential_model,
            'linear': self.linear_model,
            'gaussian': self.gaussian_model,
        }

        csv_cols = ['name', 'nugget', 'sill', 'range']
        try:
            with open(filename, 'r') as semivar_csv:
                reader = csv.DictReader(semivar_csv, fieldnames=csv_cols)
                next(reader)
                for row in reader:
                    self.params = [float(row['nugget']), float(row['sill']), float(row['range'])]
                    self.chosen_model_name = row['name']
                    try:
                        self.theoretical_model = models[self.chosen_model_name]
                    except KeyError:
                        raise KeyError('You have provided wrong model name. Available names: spherical, gaussian, '
                                       'exponential, linear.')
        except IOError:
            print("I/O error, provided path is not valid")

    def show_experimental_semivariogram(self):
        """
        Function shows experimental semivariogram of a given model.
        """
        plt.figure(figsize=(10, 10))
        plt.plot(self.empirical_semivariance[:, 0], self.empirical_semivariance[:, 1], color='blue')
        plt.title('Experimental semivariogram')
        plt.xlabel('Distance')
        plt.ylabel('Semivariance')
        plt.show()

    def show_semivariogram(self):
        """
        Function shows experimental and theoretical semivariogram in one plot.
        """
        if self.theoretical_model is None:
            raise AttributeError('Theoretical semivariogram is not calculated. \
            Did you run fit_semivariance(model_type, number_of_ranges) on your model?')
        
        x = self.calculate_values()
        plt.figure(figsize=(12, 12))
        plt.plot(self.empirical_semivariance[:, 0], self.empirical_semivariance[:, 1], color='blue')
        plt.plot(self.empirical_semivariance[:, 0], x, color='red')
        plt.legend(['Empirical semivariogram', 'Theoretical semivariogram - {} model'.format(
            self.chosen_model_name
        )])
        title_text = 'Empirical and theoretical semivariogram comparison, model error ={:.2f}'.format(
            self.model_error
        )
        plt.title(title_text)
        plt.xlabel('Distance')
        plt.ylabel('Semivariance')
        plt.show()



import numpy as np
import geopandas as gpd
import pyproj
from geopandas.tools import sjoin


def _check_columns(areal_dataframe, areal_id, points_val, points_dataframe):
    """
    Function checks if both dataframes has the same id and/or value columns to prevent program from errors.
    :param areal_dataframe: (GeoDataFrame),
    :param areal_id: (string) name of the areal id column name,
    :param points_val: (string) name of the points value column name,
    :param points_dataframe: (GeoDataFrame).
    
    :return areal_df, points_df: If areal_id column name is in points_dataframe then points_dataframe column
        name is changed with prefix pts_; If points_val column name is in areal_dataframe then areal_dataframe
        column name is changed with prefix a_. Otherwise function returns original dataframes.
    """
    
    areal_columns = areal_dataframe.columns
    point_columns = points_dataframe.columns
    
    if areal_id in point_columns:
        points_dataframe.drop(areal_id, axis=1, inplace=True)
        
    if points_val in areal_columns:
        areal_dataframe.drop(points_val, axis=1, inplace=True)
    
    return areal_dataframe, points_dataframe


def get_points_within_area(area_shapefile,
                           points_shapefile,
                           areal_id_col_name,
                           points_val_col_name,
                           dropna=True,
                           points_geometry_col_name='geometry',
                           nans_to_zero=True):
    """
    Function prepares points data for further processing.

    INPUT:

    :param area_shapefile: (string) areal data shapefile address,
    :param points_shapefile: (string) points data shapefile address,
    :param areal_id_col_name: (string) name of the column with id of areas,
    :param points_val_col_name: (string) name of the value column of each point,
    :param dropna: (bool) if True then rows with NaN are deleted (areas without any points),
    :param points_geometry_col_name: (string) default is 'geometry' as in GeoPandas GeoDataFrames,
    :param nans_to_zero: (bool) if True then all NaN values are casted to 0.

    OUTPUT:

    :return: output_points_within_area (numpy array) of area id and array with point coordinates and values
        [area_id, [point_position_x, point_position_y, value]]
    """
    output_points_within_area = []

    areal_data = gpd.read_file(area_shapefile)
    points_data = gpd.read_file(points_shapefile)

    # Test if both files have the same columns

    areal_data, points_data = _check_columns(areal_data, areal_id_col_name, points_val_col_name, points_data)

    # Test if areal data has the same projection as points data
    # Match centroid points with areas
    if not pyproj.Proj(areal_data.crs).is_exact_same(pyproj.Proj(points_data.crs)):
        points_data = points_data.to_crs(areal_data.crs)

    # Join datasets
    joined_population_points = sjoin(points_data, areal_data, how='left')

    # Drop NaN values
    if dropna:
        joined_population_points = joined_population_points.dropna(axis=0)

    # Get all areas ids
    if areal_id_col_name is None:
        areal_id_col_name = 'index_right'

    areal_ids = joined_population_points[areal_id_col_name].unique()

    # get coordinate x and coordinate y of points
    try:
        joined_population_points['x'] = joined_population_points[points_geometry_col_name].apply(
            lambda _: _.x)
        joined_population_points['y'] = joined_population_points[points_geometry_col_name].apply(
            lambda _: _.y)
    except AttributeError:
        joined_population_points['x'] = joined_population_points[points_geometry_col_name].apply(
            lambda _: _[0].x)
        joined_population_points['y'] = joined_population_points[points_geometry_col_name].apply(
            lambda _: _[0].y)

    # Set cols to hold
    cols_to_hold = ['x', 'y', points_val_col_name]

    # Get data
    for area_id in areal_ids:
        dataset = joined_population_points[joined_population_points[areal_id_col_name] == area_id]
        dataset = dataset[cols_to_hold]

        # Remove nans
        if nans_to_zero:
            dataset.fillna(0, inplace=True)

        dataset_numpy = dataset.values
        output_points_within_area.append([area_id, dataset_numpy])

    return np.array(output_points_within_area)



import unittest
import os
import geopandas as gpd
from pyinterpolate.transform.get_areal_centroids import get_centroids


class TestGetCentroids(unittest.TestCase):

    def test_get_centroids(self):
        my_dir = os.path.dirname(__file__)
        data = os.path.join(my_dir, '../sample_data/test_areas_pyinterpolate.shp')
        gdf = gpd.read_file(data)

        polygon = gdf.iloc[0]
        polygon = polygon['geometry']

        centroid = get_centroids(polygon)

        # Test length
        self.assertEqual(len(centroid), 2, "Centroid doesn't have two coordinates. Calculation error.")

        # Test values
        self.assertEqual(int(centroid[0]), 17, "Centroid calculation was wrong. Should be 17.")
        self.assertEqual(int(centroid[1]), 52, "Centroid calculation was wrong. Should be 52.")

        # Test type
        self.assertTrue(isinstance(centroid, tuple))


if __name__ == '__main__':
    unittest.main()



import numpy as np

from pyinterpolate.distance.calculate_distances import calc_point_to_point_distance
from pyinterpolate.transform.select_values_in_range import select_values_in_range


def calculate_semivariance(data, step_size, max_range):
    """Function calculates semivariance of a given set of points.

        Equation for calculation is:

            semivariance = 1 / (2 * N) * SUM(i=1, N) [z(x_i + h) - z(x_i)]^2

        where:

            - N - number of observation pairs,
            - h - distance (lag),
            - z(x_i) - value at location z_i,
            - (x_i + h) - location at a distance h from x_i.

        INPUT:

        :param data: (numpy array) coordinates and their values,
        :param step_size: (float) step size of circle within radius are analyzed points,
        :param max_range: (float) maximum range of analysis.

        OUTPUT:

        :return: semivariance: numpy array of pair of lag and semivariance values where:

            - semivariance[0] = array of lags,
            - semivariance[1] = array of lag's values,
            - semivariance[2] = array of number of points in each lag.
    """

    distances = calc_point_to_point_distance(data[:, :-1])

    semivariance = []

    lags = np.arange(0, max_range, step_size)

    # Calculate semivariances
    for h in lags:
        distances_in_range = select_values_in_range(distances, h, step_size)
        sem = (data[distances_in_range[0], 2] - data[distances_in_range[1], 2]) ** 2
        if len(sem) == 0:
            sem_value = 0
        else:
            sem_value = np.sum(sem) / (2 * len(sem))
        semivariance.append([h, sem_value, len(sem)])
    semivariance = np.vstack(semivariance)
    
    return semivariance


def calculate_weighted_semivariance(data, step_size, max_range):
    """Function calculates weighted semivariance following Monestiez et al.:

        A. Monestiez P, Dubroca L, Bonnin E, Durbec JP, Guinet C: Comparison of model based geostatistical methods
        in ecology: application to fin whale spatial distribution in northwestern Mediterranean Sea.
        In Geostatistics Banff 2004 Volume 2. Edited by: Leuangthong O, Deutsch CV. Dordrecht, The Netherlands,
        Kluwer Academic Publishers; 2005:777-786.

        B. Monestiez P, Dubroca L, Bonnin E, Durbec JP, Guinet C: Geostatistical modelling of spatial distribution
        of Balenoptera physalus in the northwestern Mediterranean Sea from sparse count data and heterogeneous
        observation efforts. Ecological Modelling 2006 in press.

        Equation for calculation is:

        s(h) = [1 / (2 * SUM(a=1, N(h)) (n(u_a) * n(u_a + h)) /...
                    / (n(u_a) + n(u_a + h)))] *...
                    * SUM(a=1, N(h)) {[(n(u_a) * n(u_a + h)) / (n(u_a) + n(u_a + h))] * [(z(u_a) - z(u_a + h))^2] - m'}

        where:

        - s(h) - Semivariogram of the risk,
        - n(u_a) - size of the population at risk in the unit a,
        - z(u_a) - mortality rate at the unit a,
        - u_a + h - area at the distance (h) from the analyzed area,
        - m' - population weighted mean of rates.

        INPUT:

        :param data: (numpy array) coordinates and their values and weights:
            [coordinate x, coordinate y, value, weight],
        :param step_size: (float) step size of circle within radius are analyzed points,
        :param max_range: (float) maximum range of analysis.


        OUTPUT:

        :return: semivariance: numpy array of pair of lag and semivariance values where:

            - semivariance[0] = array of lags
            - semivariance[1] = array of lag's values
            - semivariance[2] = array of number of points in each lag.
    """

    # TEST: test if any 0-weight is inside the dataset

    _test_weights(data[:, -1])

    # Calculate distance

    distances = calc_point_to_point_distance(data[:, :-2])

    # Prepare semivariance arrays
    smv = []
    semivariance = []

    lags = np.arange(0, max_range, step_size)

    # Calculate semivariances
    for idx, h in enumerate(lags):
        distances_in_range = select_values_in_range(distances, h, step_size)

        # Weights
        weight1 = data[distances_in_range[0], 3]
        weight2 = data[distances_in_range[1], 3]

        weights = (weight1 * weight2) / (weight1 + weight2)
        weights_sum = np.sum(weights)

        # Values
        val1 = data[distances_in_range[0], 2]
        val2 = data[distances_in_range[1], 2]

        # Weighted mean of values
        weighted_mean = ((weight1 * val1) + (weight2 * val2)) / weights_sum

        sem = weights * (data[distances_in_range[0], 2] - data[distances_in_range[1], 2]) ** 2
        sem_value = (np.sum(sem) - np.sum(weighted_mean)) / (2 * np.sum(weights_sum))
        smv.append([sem_value, len(sem)])
        if smv[idx][0] > 0:
            semivariance.append([h, smv[idx][0], smv[idx][1]])
        else:
            semivariance.append([h, 0, 0])

    semivariance = np.vstack(semivariance)
    return semivariance


def _test_weights(arr):
    if 0 in arr:
        print('One or more of weights in dataset is set to 0, this may cause errors in the distance')



import unittest
import os
from pyinterpolate.io_ops import get_points_within_area


class TestGetPoints(unittest.TestCase):

    def test_get_points(self):

        my_dir = os.path.dirname(__file__)
        areal_data = os.path.join(my_dir, '../sample_data/test_areas_pyinterpolate.shp')
        point_data = os.path.join(my_dir, '../sample_data/test_points_pyinterpolate.shp')

        points_val_col = 'value'
        areal_id_col = 'id'

        points_within_area = get_points_within_area(area_shapefile=areal_data,
                                                    points_shapefile=point_data,
                                                    areal_id_col_name=areal_id_col,
                                                    points_val_col_name=points_val_col,
                                                    dropna=True,
                                                    points_geometry_col_name='geometry',
                                                    nans_to_zero=True)

        self.assertEqual(len(points_within_area), 6, "Wrong number of points within area, should be 6")


if __name__ == '__main__':
    unittest.main()



import numpy as np
import geopandas as gpd

from pyinterpolate.transform.get_areal_centroids import get_centroids


def prepare_areal_shapefile(areal_file_address,
                            id_column_name=None,
                            value_column_name=None,
                            geometry_column_name='geometry',
                            dropnans=True):
    """Function prepares areal shapefile for processing and transforms it into numpy array. Function returns two lists.

    INPUT:

    :param areal_file_address: (string) path to the shapefile with area data,
    :param id_column_name: (string) id column name, if not provided then index column is treated as the id,
    :param value_column_name: (string) value column name, if not provided then all values are set to NaN,
    :param geometry_column_name: (string) default is 'geometry' as in GeoPandas GeoDataFrames,
    :param dropnans: (bool) if True then rows with NaN are dropped.

    OUTPUT:

    :return: areal_array (numpy array) of area id, area geometry, coordinate of centroid x, coordinate of centroid y,
        value:

        [area_id, area_geometry, centroid coordinate x, centroid coordinate y, value]
    """

    # Test if value column name is None and dropnans is True
    if (value_column_name is None) and dropnans:
        raise TypeError('You cannot leave value_column_name as None and set dropnans to True because function '
                        'will return empty list')

    shapefile = gpd.read_file(areal_file_address)
    cols_to_hold = list()

    # Prepare index column
    if id_column_name is None:
        shapefile['id_generated'] = shapefile.index
        cols_to_hold.append('id_generated')
    else:
        cols_to_hold.append(id_column_name)

    # Prepare geometry column
    cols_to_hold.append(geometry_column_name)

    # Prepare value column
    if value_column_name is None:
        shapefile['vals_generated'] = np.nan
        cols_to_hold.append('vals_generated')
    else:
        cols_to_hold.append(value_column_name)

    # Remove unwanted columns
    gdf = shapefile.copy()
    for col in gdf.columns:
        if col not in cols_to_hold:
            gdf.drop(labels=col, axis=1, inplace=True)

    # Set order of columns
    gdf = gdf[cols_to_hold]

    # Remove rows with nan's
    if dropnans:
        gdf.dropna(axis=0, inplace=True)

    # Extract values into numpy array
    areal_array = gdf.values

    # Get areal centroids
    centroids = [get_centroids(x) for x in areal_array[:, 1]]
    centroids = np.array(centroids)

    # Combine data into areal dataset
    areal_dataset = np.c_[areal_array[:, :2], centroids, areal_array[:, -1]]

    return areal_dataset



import unittest
import os
import numpy as np
from pyinterpolate.io_ops.read_data import read_point_data
from pyinterpolate.distance.calculate_distances import calc_point_to_point_distance
from pyinterpolate.semivariance.semivariogram_estimation.calculate_semivariance import calculate_semivariance
from pyinterpolate.semivariance.semivariogram_estimation.calculate_semivariance import calculate_weighted_semivariance


class TestCalculateSemivariance(unittest.TestCase):

    def test_calculate_semivariance(self):
        my_dir = os.path.dirname(__file__)
        path = os.path.join(my_dir, '../sample_data/poland_dem_gorzow_wielkopolski')

        dataset = read_point_data(path, 'txt')

        # Set semivariance params
        distances = calc_point_to_point_distance(dataset[:, :-1])

        maximum_range = np.max(distances)
        number_of_divisions = 10
        step_size = maximum_range / number_of_divisions

        gamma = calculate_semivariance(dataset, step_size, maximum_range)

        output_int = [51, 207, 416, 552, 579, 544, 517, 589, 619, 540]

        self.assertTrue((gamma[:, 1].astype(np.int) == np.array(output_int)).all(),
                        "Integer part of output should be equal to [51, 207, 416, 552, 579, 544, 517, 589, 619, 540]")


    def test_calculate_weighted_semivariance(self):
        my_dir = os.path.dirname(__file__)
        path = os.path.join(my_dir, '../sample_data/poland_dem_gorzow_wielkopolski')

        dataset = read_point_data(path, 'txt')
        new_col = np.arange(1, len(dataset) + 1)

        dataset_weights = np.zeros((dataset.shape[0], dataset.shape[1] + 1))
        dataset_weights[:, :-1] = dataset
        dataset_weights[:, -1] = new_col

        # Set semivariance params
        distances = calc_point_to_point_distance(dataset_weights[:, :-2])

        maximum_range = np.max(distances)
        number_of_divisions = 10
        step_size = maximum_range / number_of_divisions

        gamma_w = calculate_weighted_semivariance(dataset_weights, step_size, maximum_range)

        output_int = [46, 189, 384, 512, 534, 486, 438, 488, 493, 414]

        self.assertTrue((gamma_w[:, 1].astype(np.int) == np.array(output_int)).all(),
                        "Integer part of output should be equal to [46, 189, 384, 512, 534, 486, 438, 488, 493, 414]")


if __name__ == '__main__':
    unittest.main()



import numpy as np
import matplotlib.pyplot as plt

from pyinterpolate.distance.calculate_distances import calc_point_to_point_distance
from pyinterpolate.semivariance.semivariogram_estimation.calculate_semivariance import calculate_semivariance
from pyinterpolate.semivariance.semivariogram_fit.fit_semivariance import TheoreticalSemivariogram
from pyinterpolate.kriging.point_kriging.kriging import Krige


def show_data(data_matrix):
    plt.figure()
    plt.imshow(data_matrix, cmap='Spectral_r')
    plt.title('Interpolated dataset')
    plt.colorbar()
    plt.show()


def _set_dims(xs, ys, dmax):
    """
    Function sets dimensions of the output array.
    :param xs: (array) all x values,
    :param ys: (array) all y values,
    :param dmax: (int) max dimension,
    :return: x_dim_coords, y_dim_coords, [properties]
    """

    xmin = np.min(xs)
    xmax = np.max(xs)

    ymin = np.min(ys)
    ymax = np.max(ys)

    x_abs = np.abs(xmax - xmin)
    y_abs = np.abs(ymax - ymin)

    if x_abs > y_abs:
        step = x_abs / dmax
        x_dim_coords = np.arange(xmin + step, xmax + step, step)
        y_dim_coords = np.arange(ymin + step, ymax + step, step)
    else:
        step = y_abs / dmax
        y_dim_coords = np.arange(ymin + step, ymax + step, step)
        x_dim_coords = np.arange(xmin + step, xmax + step, step)

    # y_dim_coords must be flipped

    y_dim_coords = y_dim_coords[::-1]
    return x_dim_coords, y_dim_coords, [step, xmin, xmax, ymin, ymax]


def update_interpolation_matrix(rows_coords, cols_coords, kriging_model, no_of_neighbors, progress=True):
    output_vals = np.zeros(shape=(len(rows_coords), len(cols_coords)))
    output_errs = np.zeros(shape=(len(rows_coords), len(cols_coords)))

    full_size = output_vals.size
    prog = 0
    last_progress = 0
    print('Progress: {}%'.format(0))
    for ridx, point_row in enumerate(rows_coords):
        for cidx, point_col in enumerate(cols_coords):
            predicted = kriging_model.ordinary_kriging(
                [point_col, point_row], no_of_neighbors, False
            )
            output_vals[ridx, cidx] = predicted[0]
            output_errs[ridx, cidx] = predicted[1]

            if progress:

                actual_progress = int((prog / full_size) * 100)
                if actual_progress == last_progress:
                    pass
                else:
                    print('Progress: {}%'.format(actual_progress))
                    last_progress = actual_progress
                prog = prog + 1
    if progress:
        print('Progress: {}%'.format(100))

    return output_vals, output_errs


def interpolate_raster(data, dim=1000, number_of_neighbors=4, semivariogram_model=None):
    """
    Function interpolates raster from data points using ordinary kriging.

    INPUT:

    :param data: (numpy array / list) [coordinate x, coordinate y, value],
    :param dim: (int) number of pixels (points) of a larger dimension (it could be width or height),
    :param number_of_neighbors: (int) default=16, number of points used to interpolate data,
    :param semivariogram_model: (TheoreticalSemivariance) default=None, Theoretical Semivariogram model,
        if not provided then it is estimated from a given dataset.

    OUTPUT:

    :return: (numpy array) [numpy array of interpolated values, numpy array of interpolation errors,
        [pixel size, min x, max x, min y, max y]]
    """

    # Set dimension

    if isinstance(data, list):
        data = np.array(data)

    cols_coords, rows_coords, props = _set_dims(data[:, 0], data[:, 1], dim)

    # Calculate semivariance if not provided

    if semivariogram_model is None:
        distances = calc_point_to_point_distance(data[:, :-1])

        maximum_range = np.max(distances)
        number_of_divisions = 100
        step_size = maximum_range / number_of_divisions
        lags = np.arange(0, maximum_range, step_size)

        semivariance = calculate_semivariance(data, lags, step_size)

        ts = TheoreticalSemivariogram(data, semivariance, False)
        ts.find_optimal_model(False, number_of_neighbors)

    else:
        ts = semivariogram_model

    # Interpolate data point by point

    k = Krige(ts, data)

    kriged_matrix, kriged_errors = update_interpolation_matrix(rows_coords, cols_coords,
                                                               k, number_of_neighbors, True)

    return [kriged_matrix, kriged_errors], props



import re
import sys
from setuptools import setup, find_packages
from os import path


here = path.abspath(path.dirname(__file__))

# Project description
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_desc = f.read()

# obtain version string from __init__.py
with open(path.join(here, '__init__.py'), 'r') as f:
    init_py = f.read()
version = re.search("__version__ = '(.*)'", init_py).groups()[0]

setup(
    name='pyinterpolate',
    version=version,
    description='Spatial interpolation Python module',
    long_description=long_desc,
    long_description_content_type='text/markdown',
    url='https://github.com/szymon-datalions/pyinterpolate',
    download_url='https://github.com/szymon-datalions/pyinterpolate/archive/v0.2.tar.gz',
    author='Szymon Moliński',
    author_email='simon@ml-gis-service.com',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: GIS',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3.7',
    ],
    keywords=['Spatial interpolation', 'Kriging', 'Area Kriging', 'Block Kriging', 'Poisson Kriging', 'Geostatistics'],
    packages=find_packages(exclude=['data', 'test', 'tutorials', 'new_concepts']),

    install_requires=['descartes==1.1.0', 'geopandas==0.7.0', 'matplotlib==3.2.1', 'numpy==1.18.3', 'tqdm==4.47.0',
                      'pyproj==2.6.0', 'scipy==1.4.1', 'shapely==1.7.0',
		              'fiona==1.8.13.post1; sys_platform=="darwin"', 
                      'rtree==0.9.4; sys_platform=="darwin"',
		              'fiona==1.8; sys_platform=="linux"',
                      'rtree>=0.8,<0.9; sys_platform=="linux"'],
    project_urls={
	'Webpage': 'https://pyinterpolate.com',
        'Bug Reports': 'https://github.com/szymon-datalions/pyinterpolate/issues',
        'Sponsor page': 'https://datalions.eu/',
        'Source': 'https://github.com/szymon-datalions/pyinterpolate/',
    },
)



from .areal_poisson_kriging.areal_kriging import ArealKriging
from .areal_poisson_kriging.centroid_based.centroid_poisson_kriging import CentroidPoissonKriging
from .point_kriging.kriging import Krige


import unittest
import os
from pyinterpolate.semivariance.areal_semivariance.within_block_semivariance.calculate_semivariance_within_blocks\
    import calculate_semivariance_within_blocks

import numpy as np
import geopandas as gpd
from pyinterpolate.io_ops.prepare_areal_shapefile import prepare_areal_shapefile
from pyinterpolate.io_ops.get_points_within_area import get_points_within_area
from pyinterpolate.semivariance.semivariogram_estimation.calculate_semivariance import calculate_semivariance
from pyinterpolate.semivariance.semivariogram_fit.fit_semivariance import TheoreticalSemivariogram


class TestCalculateSemivarianceWithinBlocks(unittest.TestCase):

    def test_calculate_semivariance_within_blocks(self):
        # Data prepration
        my_dir = os.path.dirname(__file__)

        areal_dataset = os.path.join(my_dir, '../sample_data/test_areas_pyinterpolate.shp')
        subset = os.path.join(my_dir, '../sample_data/test_points_pyinterpolate.shp')

        a_id = 'id'
        areal_val = 'value'
        points_val = 'value'

        # Get maximum range and set step size

        gdf = gpd.read_file(areal_dataset)

        total_bounds = gdf.geometry.total_bounds
        total_bounds_x = np.abs(total_bounds[2] - total_bounds[0])
        total_bounds_y = np.abs(total_bounds[3] - total_bounds[1])

        max_range = min(total_bounds_x, total_bounds_y)
        step_size = max_range / 4

        areal_data_prepared = prepare_areal_shapefile(areal_dataset, a_id, areal_val)
        points_in_area = get_points_within_area(areal_dataset, subset, areal_id_col_name=a_id,
                                                points_val_col_name=points_val)

        # Get areal centroids with data values
        areal_centroids = areal_data_prepared[:, 2:]
        areal_centroids = np.array([[x[0], x[1], x[2]] for x in areal_centroids])

        gamma = calculate_semivariance(areal_centroids, step_size, max_range)

        # Get theoretical semivariogram model
        ts = TheoreticalSemivariogram(areal_centroids, gamma)

        ts.find_optimal_model(number_of_ranges=8)

        # Get centroids to calculate experimental semivariance

        inblock_semivariance = calculate_semivariance_within_blocks(points_in_area, ts)
        inblock_semivariance = np.array(inblock_semivariance)

        data_point = inblock_semivariance[inblock_semivariance[:, 0] == 1][0]
        self.assertEqual(int(data_point[1]), 104, "First data point's integer part should be equal to 104")


if __name__ == '__main__':
    unittest.main()



import unittest
import numpy as np
from pyinterpolate.transform.select_values_in_range import select_values_in_range


class TestValuesInRange(unittest.TestCase):

	def test_select_values_in_range(self):
		dataset = [[1, 2, 3],
	               [4, 5, 6],
	               [1, 9, 9]]
		output = (np.array([1, 1]), np.array([1, 2]))
		x = select_values_in_range(dataset, 5, 2)

		self.assertTrue((x[0] == output[0]).all(), "Output row indices should be [1, 1]")
		self.assertTrue((x[1] == output[1]).all(), "Output col indices should be [1, 2]")


if __name__ == '__main__':
	unittest.main()



import numpy as np

from pyinterpolate.kriging.areal_poisson_kriging.areal_kriging import ArealKriging
from pyinterpolate.kriging.areal_poisson_kriging.centroid_based.centroid_poisson_kriging \
    import CentroidPoissonKriging
from pyinterpolate.kriging.point_kriging.kriging import Krige


class KrigingComparison:
    """
    Class compares different areal kriging models and techniques.
    """

    def __init__(self, theoretical_semivariogram, areas, points, search_radius, ranges_of_observations,
                 training_set_frac=0.8, iters=20):
        """

        INITIALIZATION PARAMS:

        :param theoretical_semivariogram: (TheoreticalSemivariogram) modeled semivariogram,
        :param areas: (numpy array) areas in the form:
            [area_id, polygon, centroid x, centroid y, value]
        :param points: (numpy array) points within areas in the form:
            [area_id, [point_position_x, point_position_y, value]]
        :param search_radius: (float) minimal search radius to looking for neighbors,
        :param ranges_of_observations: (int) how many neighbors include in analysis,
        :param training_set_frac: (float in range 0-1) how many values set as a "known points",
        :param iters: (int) number of tests (more tests == more accurate Root Mean Squared Error of comparisons).
        """

        self.semivariance = theoretical_semivariogram
        self.areas = areas
        self.points = points
        self.radius = search_radius
        self.ranges = ranges_of_observations
        self.frac = training_set_frac
        self.iters = iters

        self.k_types = {
            'PK-ata': 0.,
            'PK-centroid': 0.,
            'OK': 0.,
            'SK': 0.
        }

        self.evaluation_output = {}  # {range: k_types and values}

    def _divide_train_test(self):
        all_ids = self.areas[:, 0]
        training_set_size = int(len(all_ids) * self.frac)
        training_ids = np.random.choice(all_ids,
                                        size=training_set_size,
                                        replace=False)
        training_areas = np.array(
            [a for a in self.areas if a[0] in training_ids]
        )
        test_areas = np.array(
            [a for a in self.areas if a[0] not in training_ids]
        )
        training_pts = np.array(
            [pt for pt in self.points if pt[0] in training_ids]
        )
        test_pts = np.array(
            [pt for pt in self.points if pt[0] not in training_ids]
        )

        output = [training_areas, training_pts,
                  test_areas, test_pts]
        return output

    def _run_pk_ata(self, training_areas, training_points, test_areas, test_points, number_of_obs):
        # Poisson Kriging model Area-to-area

        kriging_model = ArealKriging(semivariogram_model=self.semivariance,
                                     known_areas=training_areas,
                                     known_areas_points=training_points)
        pk_pred = []
        for unknown_area in test_areas:
            unknown_pts = test_points[test_points[:, 0] == unknown_area[0]][0]

            # Predict
            predicted = kriging_model.predict(unknown_pts, number_of_obs, self.radius)
            err = np.sqrt(
                (unknown_area[-1] - predicted[0])**2
            )
            pk_pred.append(err)
        return np.mean(pk_pred)

    def _run_pk_centroid(self, training_areas, training_points, test_areas, test_points, number_of_obs):
        # Poisson Kriging centroid based approach

        kriging_model = CentroidPoissonKriging(semivariogram_model=self.semivariance,
                                               known_areas=training_areas,
                                               known_areas_points=training_points)

        c_pred = []
        for unknown_area in test_areas:
            unknown_pts = test_points[test_points[:, 0] == unknown_area[0]][0]

            # Predict
            try:
                predicted = kriging_model.predict(unknown_area, unknown_pts, number_of_obs,
                                                  self.radius, True)
                err = np.sqrt(
                    (unknown_area[-1] - predicted[0]) ** 2
                )
                c_pred.append(err)
            except ValueError:
                err = unknown_area[-1]
                c_pred.append(err)
        return np.mean(c_pred)

    def _run_k_point(self, training_areas, test_areas, number_of_obs):
        # Ordinary and Simple Kriging

        kriging_data = training_areas[:, -3:]
        p_kriging = Krige(semivariogram_model=self.semivariance,
                          known_points=kriging_data)

        o_pred = []
        s_pred = []
        for unknown_area in test_areas:
            unknown_centroids = unknown_area[-3:-1]

            # Predict
            try:
                predicted = p_kriging.ordinary_kriging(unknown_centroids, number_of_obs)
                err = np.sqrt(
                    (unknown_area[-1] - predicted[0]) ** 2
                )
                o_pred.append(err)
            except ValueError:
                err = unknown_area[-1]
                o_pred.append(err)

            try:
                predicted = p_kriging.simple_kriging(unknown_centroids, number_of_obs)
                err = np.sqrt(
                    (unknown_area[-1] - predicted[0]) ** 2
                )
                s_pred.append(err)
            except ValueError:
                err = unknown_area[-1]
                s_pred.append(err)

        return np.mean(o_pred), np.mean(s_pred)

    def run_tests(self):
        """
        Method compares area-to-area, area-to-point and centroid based area Poisson Kriging.
        """
        for number_of_obs in self.ranges:
            pk_evals = []
            ck_evals = []
            ok_evals = []
            sk_evals = []

            for i in range(self.iters):
                # Generate training and test set
                sets = self._divide_train_test()

                # Poisson Kriging
                pk_eval = self._run_pk_ata(sets[0], sets[1], sets[2], sets[3], number_of_obs)
                pk_evals.append(pk_eval)

                # Centroid-based PK
                c_eval = self._run_pk_centroid(sets[0], sets[1], sets[2], sets[3], number_of_obs)
                ck_evals.append(c_eval)

                # OK and SK
                osk_eval = self._run_k_point(sets[0], sets[2], number_of_obs)
                ok_evals.append(osk_eval[0])
                sk_evals.append(osk_eval[1])

            # Mean of values

            pk_rmse = np.mean(pk_evals)
            ck_rmse = np.mean(ck_evals)
            ok_rmse = np.mean(ok_evals)
            sk_rmse = np.mean(sk_evals)

            # Update dict
            d = self.k_types.copy()

            d['PK-ata'] = float(pk_rmse)
            d['PK-centroid'] = float(ck_rmse)
            d['OK'] = float(ok_rmse)
            d['SK'] = float(sk_rmse)

            self.evaluation_output[number_of_obs] = d

            # Inform
            print('Evaluation metrics updated')
            print('Number of ranges:', number_of_obs)
            print('ROOT MEAN SQUARED ERROR VALUES')
            print('Poisson Kriging ATA -', pk_rmse)
            print('Poisson Kriging centroids -', ck_rmse)
            print('Ordinary Kriging -', ok_rmse)
            print('Simple Kriging -', sk_rmse)



def get_centroids(polygon):
    """
    Function gets position x and y of polygon centroid.

    INPUT:

    :param polygon: (shapely Polygon).

    OUTPUT:

    :return: (tuple) centroid position (x, y) for a given area.
    """

    centroid_position_x = polygon.centroid.x
    centroid_position_y = polygon.centroid.y

    return centroid_position_x, centroid_position_y
